import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.zip.DataFormatException;

import org.jdom2.Document;
import org.jdom2.Element;
import org.jdom2.JDOMException;
import org.jdom2.input.SAXBuilder;


public class CoordTransformer
{
	private boolean _initialized;
	private int _inputNodes;
	private String _inputMap;
	private List<FileInputStream> _inputFiles;
	private ArrayList<Street> _streets;
	private ArrayList<Node> _nodes;
	private FileWriter _outputFile;
	
	public CoordTransformer(int nodes, String map)
	{
		_inputNodes = nodes;
		_inputMap = map;
		_initialized = false;
		_streets = new ArrayList<Street>();
		_nodes = new ArrayList<Node>();
		//create our file input stream list
		_inputFiles = new ArrayList<FileInputStream>();
	}
	
	public void Close()
	{

		try
		{
			for(FileInputStream fs : _inputFiles)
			{
				fs.close();
			}
			_outputFile.close();
		}
		catch(Exception e)
		{
			System.out.println(e.getMessage());
		}
	}
	
	public boolean Initialize()
	{
		boolean result = true;
		try
		{
			if(_inputMap == "")
			{
				throw new DataFormatException("No input map specified. Please specifie a map with the -f option.");
			}
			if(_inputNodes == 0)
			{
				throw new DataFormatException("No or invalid number of nodes given. Please set nodes with the -n option.");
			}
			//set capacity to Number of Input files + 2 Files (Nodes + Streets)
			((ArrayList<FileInputStream>)_inputFiles).ensureCapacity(_inputNodes + 2);
			for(int i = 0; i < _inputNodes; i++)
			{
				//open each node file car_dump_MPIRANK.dat
				_inputFiles.set(i, new FileInputStream("car_dump_" + i + ".dat"));
			}
			//open data files
			_inputFiles.set(_inputNodes, new FileInputStream("MapData/" + _inputMap + "/nodes_" + _inputMap + "_tmp.dat"));
			_inputFiles.set(_inputNodes + 1, new FileInputStream("MapData/" + _inputMap + "/fullways_" + _inputMap + "_tmp.xml"));
			
			//if we were able to read our files, try to read the Map files
			readNodes();
			readStreets();
			
			//initialize output file
			File ausgabe = new File("ausgabe_" + _inputMap + ".dat");
			_outputFile =  new FileWriter(ausgabe, false);
		}
		catch(Exception e)
		{
			System.out.println(e.getMessage());
			result = false;
		}
		_initialized = result;
		return result;
	}
	
	
	private void readStreets() throws FileNotFoundException
	{	
		FileReader fr = new FileReader("MapData/" + _inputMap + "/fullways_" + _inputMap + "_tmp.xml");
		SAXBuilder sxb = new SAXBuilder();
		Document document;
		try
		{
			document = sxb.build(fr);
			// Wurzelelement
			Element element = document.getRootElement();
			// Alle Wege
			List<Element> list = element.getChildren();
			//ber jeden Weg iterrieren
			for(Element weg : list)
			{		
				ArrayList<Pair<Node,Double>> nodes = new ArrayList<Pair<Node,Double>>();
				long id = Long.parseLong(weg.getChild("id").getText());
				List<Element> knoten = weg.getChild("Nodes").getChildren();
				for (Element node : knoten)
				{
					Pair<Node, Double> n = 
							new Pair<Node,Double>(_nodes.get(Integer.parseInt(node.getAttributeValue("id"))),
									Double.parseDouble(node.getText()));
					nodes.add(n);
				}
				Street street = new Street(id, nodes);
				_streets.add(street);
			}

		} catch (JDOMException e)
		{
			e.printStackTrace();
		} catch (IOException e)
		{
			e.printStackTrace();
		}
	}
	
	public boolean processNext()
	{
		boolean result = true;
		//check input files
		try
		{
			for(int i = 0; i < _inputNodes; i++)
				if(_inputFiles.get(i).available() < 16)
				{
					result = false;
					break;
				}
			if(result)
			{
				List<Pair<Integer,Double>> values = new LinkedList<Pair<Integer,Double>>();
				for(int i = 0; i < _inputNodes; i++)
				{
					readTimeStep(values, i);
				}
				//we read our data from all files, now process'em
				processTimeStep(values);
			}
		}
		catch(Exception e)
		{
			System.out.println(e.getMessage());
			result = false;
		}
		return result;
	}
	
	private void readTimeStep(List<Pair<Integer,Double>> values, int nodeID) throws IOException
	{
		//file byte buffer [4 Byte Integer (32Bit)]
		byte[] b = new byte[4];
		_inputFiles.get(nodeID).read(b);
		//convert byte Buffer to Int (Little Endian)
		int iAnzPairs = byteArrayToInt(b);
		for (int i = 0; i < iAnzPairs; i++)
		{
			//read Sreeet ID (32 Bit Integer)
			byte[] a = new byte[4];
			_inputFiles.get(nodeID).read(a);
			
			//read distance (Double 64 Bit)
			byte[] c = new byte[8];
			_inputFiles.get(nodeID).read(c);
			int id = byteArrayToInt(a);
			values.add(new Pair<Integer, Double>(id,toDouble(c)));
		}
	}
	
	private void processTimeStep(List<Pair<Integer, Double>> values) throws IOException
	{
		_outputFile.write(values.size()+ "\r\n");
		//go over every value pair of our current timestamp
		for (Pair<Integer, Double> valuePair : values)
		{
			Street s = _streets.get(valuePair.getLeft()); 
			double traveledDistance = valuePair.getRight(); // die lnge die sich das
													// auto
													// auf der gesamten
													// Strae
													// bewegt hat.
			// Ruft mit der Straenid und der Lnge ab, zwischen welchen 2
			// Knotenpunkten der Strae sich das Auto befindet. In nodes
			// werden die KnotenID und die Lngen der beiden Knoten jeweils
			// als Double zurckgegeben. Die Variablen mit "2" sind der
			// jeweils grere knoten.
			Pair<Node,Double> firstNode = s.kleinerKnoten(traveledDistance);
			Pair<Node,Double> secondNode = s.groerKnoten(traveledDistance);
			// autolaenge: autolaenge die das Auto auf der Strae unterwegs
			// ist.
			// kleinX: 1. Koordinate des 1. Knotens
			// kleinY: 2. Koordinate des 1. Knotens
			// kleinlnge: abstand des 1. knotens zum 0 knoten.
			// Fr 2. Knoten statt klein, gro.
			double kleinx = firstNode.getLeft()._x;
			double kleiny = firstNode.getLeft()._y;
			double kleinlaenge = firstNode.getRight();
			double grox = secondNode.getLeft()._x;
			double groy = secondNode.getLeft()._y;
			double grolaenge = secondNode.getRight();
			double knotenabstand = grolaenge - kleinlaenge;
			double autolnge = traveledDistance - kleinlaenge;
			double fortschrittprozent = (autolnge / knotenabstand);
			double autokoordinatex = (((grox - kleinx) * fortschrittprozent) + kleinx);
			double autokoordinatey = (((groy - kleiny) * fortschrittprozent) + kleiny);
			if (!(String.valueOf(autokoordinatey).equals("NaN")
					|| String.valueOf(autokoordinatey).equals("NaN")))
			{
				_outputFile.write(String.valueOf(autokoordinatex) + ";");
				_outputFile.write(String.valueOf(autokoordinatey));
				_outputFile.write("\r\n");
			}
		}
	}
	
	private void readNodes() throws FileNotFoundException
	{
		FileReader fr = new FileReader("MapData/" + _inputMap + "/nodes_" + _inputMap + "_tmp.dat");
		BufferedReader br = new BufferedReader(fr);
		String s;
		try
		{
			s = br.readLine();
			//node count
			s = br.readLine();
			int nodeCount = Integer.parseInt(s);
			_nodes.ensureCapacity(nodeCount);
			s = br.readLine();
			while ((s = br.readLine()) != null)
			{
				String[] nodes = s.split(";");
				Node d = new Node(Integer.parseInt(nodes[0]),
						Double.parseDouble(nodes[1]),
						Double.parseDouble(nodes[2]));
				_nodes.set(d._id, d);
			}
			fr.close();
			// Liste mit Zeilenweise der knoten-koordinaten datei.
		} catch (IOException e)
		{

			e.printStackTrace();
		}
	}
	
	/**
	 * Wandelt ein Byte-Array (Lnge 4 Bytes) in einen Long-Wert um.
	 * 
	 * @param b
	 * @return
	 */
	public int byteArrayToInt(byte[] b)
	{
		int value = 0;
		for (int i = 0; i < b.length; i++)
		{
			value += ((int) b[i] & 0xffL) << (8 * i);
		}
		return value;
	}
	

	/**
	 * Wandelt ein Byte-Array (Lnge 8 Bytes) in einen Double-Wert um.
	 * 
	 * @param bytes
	 * @return
	 */
	public double toDouble(byte[] bytes)
	{
		double d = ByteBuffer.wrap(bytes).order(ByteOrder.LITTLE_ENDIAN)
				.getDouble();
		return d;
	}
}
