/*
Serial traffic simulation program for PaPo.

############################################

This application uses data from Open Street Map, henceforth
that data is made available under the Open Database License: http://opendatacommons.org/licenses/odbl/1.0/.
Any rights in individual contents ofthe database are licensed under 
the Database Contents License: http://opendatacommons.org/licenses/dbcl/1.0/
See more at: http://opendatacommons.org/licenses/odbl/#sthash.y0L9Yy1N.dpuf

############################################

Author: Martin Poppinga, Miriam von Heeremann, Marcel Ellermann
Date: 24.07.2013      
Distributed under the Boost Software License, Version 1.0.
See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)

#############################################
*/

#include "Car.h"
#include "Log.h"
#include "Street.h"
#include "MapHandler.h"
#include "SpawnHandler.h"
#include "boost/format.hpp"
#include "ConfigHandler.h"

namespace traffic_simulation
{
	Car::Car(ui64 guid, Node* n, Street* s, std::list<std::pair<Node*, Street*>*>* path) : WorldObj(n->getPos())
	{
		//set current goal node to our spawn point
		//used in initPath to generate a path
		_cgNode.first = n;
		_cgNode.second = s;
		_cStreet = s;
		_guid = guid;
		_cTimeStamp = WORLD_TIME_STAMP + 1;
		_speed = 0;
		_wPriority = false;
		_notified = false;
		_cPath = NULL;
		setPath(path);
		if(_cStreet != _cgNode.second)
			_cDirection = (*_cStreet->nodes()->begin()) == _cgNode.first ? driveDirection::backward : driveDirection::forward;
		else
		{
			for(std::list<Node*>::const_iterator iNode = _cStreet->nodes()->begin(); iNode != _cStreet->nodes()->end(); iNode++)
			{
				if((*iNode)->getID() == _cgNode.first->getID())
				{
					_cDirection = driveDirection::backward;
					break;
				}
				else if((*iNode)->getID() == n->getID())
				{
					_cDirection = driveDirection::forward;
					break;
				}
			}
		}
		_cMePos = _cStreet->getNodePosition(n,_cDirection);
		perfomNodeLocks();
	}

	Car::~Car()
	{
		//free allocated memory
		for(std::list<std::pair<Node*, Street*>*>::iterator itr = _cPath->begin(); itr != _cPath->end(); itr++)
			delete(*itr);
		delete(_cPath);
		perfomNodeLocks(true); //remove all node locks
	}

	/// <summary>
	/// Sets the path for the car.
	/// </summary>
	/// <param name="path">The path. Be sure to not pass an empty path</param>
	void Car::setPath(std::list<std::pair<Node*, Street*>*>* path)
	{
		//free allocated memory from path
		if(_cPath != NULL)
		{
			for(std::list<std::pair<Node*, Street*>*>::iterator itr = _cPath->begin(); itr != _cPath->end(); itr++)
				delete(*itr);
			delete(_cPath);
		}
		_cPath = path;
		_cgNode = **path->begin();
	}

	/// <summary>
	/// Acceleration to have security distance to dist
	/// </summary>
	/// <param name="dist">The dist.</param>
	/// <param name="speed">The speed.</param>
	/// <returns></returns>
	double __accToBreakPoint(double dist, double cSpeed, double dstSpeed)
	{
		if(dist <= 0)
			return dstSpeed - cSpeed;
		double _r = (pow(dstSpeed,2)-pow(cSpeed,2))/(2*dist);
		return _r < EM_N_ACCELERATION ? EM_N_ACCELERATION : _r;
	}

	///<summar>
	/// Method to handle acceleration of our car.
	///</summar>
	double Car::regulateSpeed()
	{
		///ToDo: brake before streets with cars that may have prioity
		///ToDO: implement traffic lights
		///General behaviour -- slow down if distance to front car is to less
		///					 -- brake if we can't pass intersection
		// default acceleration, now check all possible acceleration modifiers
		double acceleration = ACCELERATION;
		double tmpAcc = 0;
		double eDist = 0;

		//get all streets we have to care for in the next 3 seconds
		bool eReached = false;
		double chkSpeed = std::max(_speed, 2 * SEC_TO_LOGIC_TIME);
		std::list<std::pair<Node*,Street*>> lStreets = getReqStreetsForDist(chkSpeed * 3 * SEC_TO_LOGIC_TIME, &eReached);
		for(std::list<std::pair<Node*,Street*>>::iterator iStreet = lStreets.begin(); iStreet != lStreets.end(); iStreet++)
		{
			Street* street = iStreet->second;
			Node* node = iStreet->first;
			//get distance to goal node
			//lenght of street + distance on old street
			eDist = street->length() + eDist;
			//if street is our current street, subtract the distance we already have moved on our street
			if(street == _cStreet)
				eDist -= _cMePos;
			//get next street and next goal node
			std::list<std::pair<Node*,Street*>>::iterator nIStreet = std::next(iStreet);
			Street* nStreet = nIStreet != lStreets.end() ? nIStreet->second : NULL;
			Node* nNode = nIStreet != lStreets.end() ? nIStreet->first : NULL;
			//distance required to slow down before priority streets
			double eBDist = PRIORITY_STREET_PASS_SPEED * LOGIC_TIME_TO_SEC;
			//end point behaviour
			if(nStreet != NULL)
			{
				street->notifyCloseToEnd(_guid, _cgNode.first->getID(), _cgNode.second->id());
				//adjust speed if we're going to drive on a street with lower maximum speed than our current speed
				if(nStreet->speed() < street->speed() && eDist < (pow(nStreet->speed(),2)-pow(_speed,2)) / (2*N_ACCELERATION))
				{
					acceleration = __accToBreakPoint(eDist, _speed, nStreet->speed());
				}


				if((eDist < street->speed() * 3 * SEC_TO_LOGIC_TIME && !nStreet->canChangeOnStreet(nNode->getID()))
					|| _wPriority)
				{
					tmpAcc = __accToBreakPoint(eDist, _speed, 0);
					acceleration = tmpAcc < acceleration ? tmpAcc : acceleration;
				}

				//comput current direction
				driveDirection sDirection = _cDirection;
				if(street != nStreet)
					sDirection = (*nStreet->nodes()->begin()) == nNode ? driveDirection::forward : driveDirection::backward;
				else if(node != NULL)
					sDirection = (*street->nodes()->begin()) == node ? driveDirection::forward : driveDirection::backward;

				if((eDist < (pow(street->speed(),2))/N_ACCELERATION) && street->hasTrafficLightLock(sDirection))
				{
					tmpAcc = __accToBreakPoint(eDist - 1,_speed, 0);
					if(tmpAcc > EM_N_ACCELERATION && tmpAcc < acceleration)
						acceleration = tmpAcc;
				}
				//don't cross intersections with max speed if we might have to give priority to another street
				if(!street->isPriorityStreet(nStreet))
				{
					eDist -= eBDist; //set distance to the point where we have to be at 20km/h -- so 1 sec before the goal node -> -1*5.555 m/s
					double minDist = (pow(PRIORITY_STREET_PASS_SPEED,2)-pow(_speed,2)) / (2*N_ACCELERATION);
					if(minDist >= eDist)
					{
						tmpAcc = __accToBreakPoint(eDist, _speed, PRIORITY_STREET_PASS_SPEED);
						acceleration = tmpAcc < acceleration ? tmpAcc : acceleration;
					}
				}
			}
			else
			{
				_cStreet->notifyNotCloseToBegin(_guid);
				_notified = false;
			}
			if(_cStreet == street)
			{
				//check safety distance to front car
				if(_fCar != NULL)
				{
					double dist = _fCar->getMetricPos() - _cMePos - CAR_LENGTH;
					double bDist = dist - 1; //stop 1 m behind the car

					if(dist <= 0)
					{
#ifdef TF_DEBUG_MAIN
						boost::format fmt = boost::format("Car: %1% crashed with car: %2% on street %3%, distance: %4%") % this->_guid % _fCar->getGUID() % _cStreet->id() % dist;
						LOG->writeTrafficLog(_FMT_TO_C_STR_(fmt));
#endif
					} 

					else if((bDist) <= (-pow((_speed),2)/(N_ACCELERATION)) || bDist <= 0.5 || bDist < _speed)//stop 1 m behind the car
					{
						tmpAcc = __accToBreakPoint(bDist, _speed, 0);
						acceleration = tmpAcc < acceleration ? tmpAcc : acceleration;
					}
					else if(acceleration > 0 && _speed < _fCar->getCurrentSpeed() && bDist <= _cStreet->speed() * 2 * SEC_TO_LOGIC_TIME)
					{
						tmpAcc = _fCar->getCurrentSpeed() - _speed;
						acceleration = tmpAcc > ACCELERATION ? ACCELERATION : tmpAcc;
					}
					else if(acceleration > 0 && _speed >= _fCar->getCurrentSpeed() && bDist <= _cStreet->speed() * 2 * SEC_TO_LOGIC_TIME)
					{
						acceleration = 0;
					}
				}
			}
		}
		//set new speed
		//don't accelerate if we're already at max speed
		if(_speed == (double)_cStreet->speed() && acceleration >= 0)
		{
			acceleration = 0;
		}
		else
		{
			//we can't drive backwards
			if(acceleration + _speed < 0)
			{
				acceleration = -_speed;
			}
			else if(acceleration + _speed > _cStreet->speed()) //we can't drive faster than the maximum speed
			{
				acceleration = _cStreet->speed() - _speed;
			}
		}
		_speed += acceleration; //apply acceleration
		if(_speed < 0.1) //just stop if we're too slow
		{
			acceleration -= _speed;
			_speed = 0;
		}
		return acceleration;
	}

	void Car::performMovement(double acc)
	{
		if(_cPath != NULL && _cPath->size() > 0)
		{
			//s = 0.5 * a * t + v0 * t
			//t = 1
			double travelDistance = acc * 0.5 + (_speed - acc); //v0 = is our old speed
			bool reachEnd = false;
			std::list<std::pair<Node*,Street*>> streets = getReqStreetsForDist(travelDistance, &reachEnd);
			if(reachEnd)
			{
				SpawnHandler::getInstance()->notifyPathEndReached(this);
				return;
			}
			double eDist = _cStreet->length() - _cMePos;
			ui32 oldStreet = _cStreet->id();
			ui32 newStreet = oldStreet;
			ui32 newNode = 0;
			while(eDist <= travelDistance && !_wPriority && _cgNode.second->canChangeOnStreet(_cgNode.first->getID()) && _fCar == NULL
				&& (!_cStreet->hasTrafficLightLock(_cDirection) || (_cStreet->hasTrafficLightLock(_cDirection) && _speed > 1*SEC_TO_LOGIC_TIME)))
			{
				travelDistance -= eDist;
				perfomNodeLocks(true); //remove all locks. new locks are applied after we are on our next street.
				//update drive direction here cauz it's depends on entry node of new street
				if(_cgNode.first == (*_cgNode.second->nodes()->begin()))
					_cDirection = driveDirection::forward;
				else
					_cDirection = driveDirection::backward;
				_cStreet->notifyNotCloseToEnd(_guid);
				_cStreet = _cgNode.second;
				newStreet = _cStreet->id();
				newNode = _cgNode.first->getID();
				delete(*_cPath->begin());
				_cPath->erase(_cPath->begin());
				_cgNode = **_cPath->begin();
				_cMePos = 0;
				eDist = _cStreet->length();
#ifdef TF_DEBUG_DETAIL_FULL
				boost::format fmt = boost::format("Car: %1%, changing on street: %2%, driving forward: %3%") % this->_guid % _cStreet->id() % (_cDirection == driveDirection::forward);
				LOG->writeCarLog(_FMT_TO_C_STR_(fmt));
#endif		
			}
#ifdef TF_DEBUG_DETAIL_FULL
			if (!_cgNode.second->canChangeOnStreet(_cgNode.first))
			{

				boost::format fmt = boost::format("Car: %1%, waiting for Street: %2% to be free at Node: %3%...") % this->_guid % _cStreet->id() % _cgNode.first->getID();
				LOG->writeCarLog(_FMT_TO_C_STR_(fmt));
			}
#endif
			_cMePos = (eDist > travelDistance) ? _cMePos + travelDistance : _cMePos + eDist;
			//update end distance after we made our move and notify if neccesary
			eDist = _cStreet->length() - _cMePos;
			//Let's assume that if we are on a priority street, all non priority streets have a well implemented behaviour, that the next street always is free.
			if(oldStreet != newStreet)
			{
				MapHandler::getInstance()->changeStreet(this, oldStreet, newNode, newStreet);
				_cStreet->notifyCloseToBegin(_guid, newNode);
			}
			else if(eDist <= _cStreet->speed() * 2 * SEC_TO_LOGIC_TIME)
			{
				_cStreet->notifyCloseToEnd(_guid, _cgNode.first->getID(), _cgNode.second->id());
			}

		}
#ifdef TF_DEBUG_MAIN
		else
		{
			boost::format fmt = boost::format("Car: %1%, has no valid path to move along.") % this->_guid;
			LOG->writeCarLog(_FMT_TO_C_STR_(fmt));
		}
#endif
	}


	/// <summary>
	/// Gets the streets that are requried for speed / travel compuation for current speed.
	/// </summary>
	/// <param name="reachEnd">Boolean value that is set to true if the car reaches the end at the current speed.</param>
	/// <returns>
	/// A list of required streets
	/// </returns>
	std::list<std::pair<Node*,Street*>> Car::getReqStreetsForDist(double distance, bool *reachEnd)
	{
		std::list<std::pair<Node*,Street*>> returnSet;
		//add current street to list, we always have to care for our current street
		returnSet.push_back(std::pair<Node*, Street*>((Node*)NULL, _cStreet));
		distance -= _cStreet->length() - _cMePos; //distance left on this street
		std::list<std::pair<Node*,Street*>*>::iterator cGPair = _cPath->begin();
		while(distance >= 0 && cGPair != _cPath->end())
		{
			returnSet.push_back(**cGPair);
			distance -= (*cGPair)->second->length();
			cGPair++;
		}
		*reachEnd = (cGPair == _cPath->end());
		return returnSet;
	}


	/// <summary>
	/// Checks for priority streets.
	/// Sets _wPriority to true if there is one priority street in our intersection list, that has a car close to its end.
	/// </summary>
	void Car::checkForPriorityStreet()
	{
		_wPriority = false;
		//check if we have to care for any streets at the next intersection
		std::pair<Street::IntersectionMap::const_iterator, Street::IntersectionMap::const_iterator> d = _cStreet->intersections()->equal_range(_cgNode.first->getID());
		for(Street::IntersectionMap::const_iterator itr = d.first; itr != d.second; itr++)
		{
			if(itr->second.first == _cgNode.second) //found our street
			{
				for(std::list<Street*>::const_iterator st = itr->second.second.cbegin(); st != itr->second.second.cend(); st++)
				{
					if((*st)->hasPriorityCar(_cgNode.first->getID()))
					{
						//check if piority cars lock'em selves..
						std::set<Street*> closedList;
						_wPriority = (!isPriorityLocked(_cStreet->id(), _cgNode.second->id(), _cgNode.first->getID(), closedList));
						//if priority is locked, just let the car ignore the piority rules and solve the lock
						if(!_wPriority)
							return;
					}
				}
			}
		}
		return;
	}

	bool Car::isPriorityLocked(ui32 cStreet, ui32 gStreet, ui32 gNode, std::set<Street*> &closedList) const
	{
		std::set<Street*> prioStreets = _cStreet->getPriorityStreets(gNode, gStreet);
		for(std::set<Street*>::iterator iPrio = prioStreets.begin(); iPrio != prioStreets.end(); iPrio++)
		{
			if(closedList.find(*iPrio) == closedList.end())
			{
				closedList.insert(*iPrio);
				std::list<std::tuple<ui64,ui32,ui32>> pCars = (*iPrio)->getPriorityCars(gNode);
				if(!pCars.empty())
				{
					if(isPriorityLocked((*iPrio)->id(), std::get<1>(*(pCars.begin())), std::get<2>(*pCars.begin()), closedList))
					{
						return true;
					}
				}
			}
			else
			{
				return((*iPrio)->id() == (*closedList.begin())->id());
			}
		}
		return false;
	}


	/// <summary>
	/// Locks and unlocks nodes.
	/// Lock nodes that are in immediate distance to our car, about 7secs max driving speed
	/// Unlock nodes that we passed. Distance < -CAR_LENGTH
	/// </summary>
	/// <param name="freeAll">True if no locks should be perfomed and all nodes should be unlocked.</param>
	void Car::perfomNodeLocks(bool freeAll)
	{
		//unlock all nodes
		if(freeAll)
		{
			for(std::list<ui32>::iterator nodeID = _locks.begin(); nodeID != _locks.end();)
			{
				NODES->removeNodeLock(*nodeID, this->_guid); //remove lock on node
				nodeID = _locks.erase(nodeID);
			}
		}
		else //lock close nodes and unlock passed nodes
		{
			//lock nodes
			for(std::list<Node*>::const_iterator n = _cStreet->nodes()->begin(); n != _cStreet->nodes()->end(); n++)
			{
				double dist = _cStreet->getNodePosition(*n, _cDirection) - _cMePos; //get distance between car and node
				if(dist >= NODE_LOCK_DISTANCE_EM_B && dist < NODE_LOCK_TIME_LT_F * _cStreet->speed())
				{
					bool lock = true;
					//check of we already locked this node
					for(std::list<ui32>::iterator nodeID = _locks.begin(); nodeID != _locks.end(); nodeID++)
					{
						if(*nodeID == (*n)->getID())
							lock = false;
					}
					if(lock)
					{
						_locks.push_back((*n)->getID());
						NODES->addNodeLock((*n)->getID(), _guid, _cDirection);
					}
				}
			}
			//free nodes
			for(std::list<ui32>::iterator nodeID = _locks.begin(); nodeID != _locks.end();)
			{
				double dist = _cStreet->getNodePosition(NODES->getNode(*nodeID), _cDirection) - _cMePos; //get distance between car and node
				if(dist < NODE_LOCK_DISTANCE_EM_B)
				{
					NODES->removeNodeLock(*nodeID,_guid);
					nodeID = _locks.erase(nodeID);
				}
				else
					nodeID++;
			}
		}
	}

	void Car::update(ui8 step)
	{
		//boost::format frmt = boost::format("Update Car: %1%") % _guid;
		//LOG->writeCarLog(boost::str(frmt).c_str());
		//do movement code here for the car
		//get car in front of this one
		_cTimeStamp++; //increase time stamp -- no more updates for this car in this this update cycle
		_fCar = _cStreet->getFrontCar(this);
		double eDist = _cStreet->length() - _cMePos;
		if(eDist < _cStreet->speed() * 2 * SEC_TO_LOGIC_TIME)
			checkForPriorityStreet(); //check if we have any priority streets to care of
		double acc = regulateSpeed();
		performMovement(acc);
		perfomNodeLocks();
	}
}
