/*
Serial traffic simulation program for PaPo.

############################################

This application uses data from Open Street Map, henceforth
that data is made available under the Open Database License: http://opendatacommons.org/licenses/odbl/1.0/.
Any rights in individual contents ofthe database are licensed under 
the Database Contents License: http://opendatacommons.org/licenses/dbcl/1.0/
See more at: http://opendatacommons.org/licenses/odbl/#sthash.y0L9Yy1N.dpuf

############################################

Author: Martin Poppinga, Miriam von Heeremann, Marcel Ellermann
Date: 24.07.2013      
Distributed under the Boost Software License, Version 1.0.
See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)

#############################################
*/

#include "CarHandler.h"
#include "Street.h"
#include "Log.h"
#include "boost/format.hpp"
#include "ConfigHandler.h"
#include "MessageHandler.h"
#include <fstream>
#include <mpi.h>

namespace traffic_simulation
{

	/// <summary>
	/// Initializes a new instance of the <see cref="CarHandler"/> class.
	/// <see cref="CarHandler" /> is responsible for creating, managing and removing cars from the current world.
	/// </summary>
	CarHandler::CarHandler()
	{
		_globalCarCount = 0;
#ifdef CAR_DUMP
		boost::format fmt = boost::format("%1%_%2%.dat") % PATH_CAR_DUMP % world::getCommRank(); 
		_carDumpFile = new std::ofstream(_FMT_TO_C_STR_(fmt), std::ios::binary);
#endif
	}

	CarHandler::~CarHandler()
	{
		//cleanup our allocated memory
		for(std::list<Car*>::iterator itr = _lCars.begin(); itr != _lCars.end(); itr++)
		{
			//free every created car
			free(*itr);
		}
#ifdef CAR_DUMP
		_carDumpFile->close();
		delete(_carDumpFile);
#endif
	}

	void CarHandler::updateCarCount()
	{
			ui32 localCarCount = _lCars.size();
			MPI_Allreduce(&localCarCount, &_globalCarCount, 1, MPI_UNSIGNED_LONG, MPI_SUM, MPI_COMM_WORLD);
	}

	Car* CarHandler::addCarToWorld(ui64 guid, Node* n, Street* s, std::list<std::pair<Node*, Street*>*>*  path)
	{
		if(path != NULL && path->size() > 0)
		{
			Car* c = new Car(guid, n, s, path);
			if(s->spawnCar(c, c->getDirection()))
			{
#ifdef TF_DEBUG_DETAIL
				boost::format fmt = boost::format("Adding Car: %1% to world (N: %2%, S: %3%)") % guid % n->getID() % s->id();
				LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
#endif
#pragma omp critical (addCarToList)
				{
					_lCars.push_back(c);
				}
				return c;
			} //unable to spawn car. delete it.
			else
				delete(c);
		}
#ifdef TF_DEBUG_MAIN
		else
		{
			boost::format fmt = boost::format("Invalid path given for Car: %1%") % guid;
			LOG->writeCarLog(_FMT_TO_C_STR_(fmt));
		}
#endif
		return NULL;
	}

	/// <summary>
	/// Returns the current number of cars handled by the car Handler
	/// </summary>
	/// <returns>Number of cars</returns>
	ui32 CarHandler::carCount()
	{
		return _globalCarCount;
	}

	void CarHandler::removeCarFromWorld(Car* c)
	{
		c->getCurrentStreet()->removeCar(c);
		_lCars.remove(c);
		delete(c);
	}

	void CarHandler::dumpCars()
	{
#ifdef CAR_DUMP
		if(*_carDumpFile)
		{
			ui32  size = _lCars.size();
			_carDumpFile->write((char*)&size,sizeof(size));
			for(std::list<Car*>::iterator iCar = _lCars.begin(); iCar != _lCars.end(); iCar++)
			{
				ui32 id = (ui32)(*iCar)->getCurrentStreet()->id();
				bool forward = (*iCar)->getDirection() == driveDirection::forward;
				double distance = forward ? (*iCar)->getMetricPos() : (*iCar)->getCurrentStreet()->length() - (*iCar)->getMetricPos();
				if(distance > (*iCar)->getCurrentStreet()->length())
				{
					boost::format fmt = boost::format("Car: %1% with metric position %2% on street %3%, which total lengh is %4%")
						% (*iCar)->getGUID() % (*iCar)->getMetricPos() % id % (*iCar)->getCurrentStreet()->length();
					LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
				}
				_carDumpFile->write((char*)&id,sizeof(id));
				_carDumpFile->write((char*)&distance, sizeof(distance));
				_carDumpFile->flush();
			}
		}
#endif
	}
}

