/*
Serial traffic simulation program for PaPo.

############################################

This application uses data from Open Street Map, henceforth
that data is made available under the Open Database License: http://opendatacommons.org/licenses/odbl/1.0/.
Any rights in individual contents ofthe database are licensed under 
the Database Contents License: http://opendatacommons.org/licenses/dbcl/1.0/
See more at: http://opendatacommons.org/licenses/odbl/#sthash.y0L9Yy1N.dpuf

############################################

Author: Martin Poppinga, Miriam von Heeremann, Marcel Ellermann
Date: 24.07.2013      
Distributed under the Boost Software License, Version 1.0.
See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)

#############################################
*/

#include "ConfigHandler.h"
#include "boost/format.hpp"
#include "Log.h"
#include <iostream>
#include <sstream>


namespace traffic_simulation
{
	bool ConfigHandler::_bInstance = false;
	bool ConfigHandler::_bLoaded = false;
	ConfigHandler* ConfigHandler::_instance = NULL;



	/// <summary>
	/// Returns the instance of our <see cref="ConfigHandler" /> class to secure that only a single instance is created.
	/// </summary>
	/// <returns></returns>
	ConfigHandler* ConfigHandler::getInstance()
	{
		if(!_bInstance)
		{
			_instance = new ConfigHandler();
			_bInstance = true;
		}
		return _instance;
	}

	/// <summary>
	/// Prevents a default instance of the <see cref="NodeHandler"/> class from being created.
	/// </summary>
	ConfigHandler::ConfigHandler()
	{
		_bLoaded = loadConfigFile();
	}

	/// <summary>
	/// Finalizes an instance of the <see cref="NodeHandler"/> class.
	/// </summary>
	ConfigHandler::~ConfigHandler()
	{
		_bInstance = false;
		_instance = NULL;
		_bLoaded = false;
	}

	bool ConfigHandler::loadConfigFile()
	{
		bool result = true;
		std::ifstream in("Simulation.config");
		LOG->writeWorldLog("Start loading Config File");
		if(in)
		{
			std::string inputLine;
			int line = 1;
			while(std::getline(in, inputLine))
			{
				std::stringstream sline(inputLine);
				std::string key, value;
				std::getline(sline, key, '=');
				std::getline(sline, value, '=');
				if(key != "")
				{
					if(value != "")
					{
						_confValues.insert(std::make_pair(key,value));
					}
					else
					{
						boost::format fmt = boost::format("Error: Empty value value specified in Line %1% for Key: %2%") % line % key;
						LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
						result = false;
						break;
					}
				}
				else
				{
					boost::format fmt = boost::format("Error: Empty key value specified in Line %1%") % line;
					LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
					result = false;
					break;
				}
				line++;
			}
		}
		else
		{
			LOG->writeWorldLog("Couldn't find Config File. Make sure Simulation.config is in the same direction as your simulation executable.");
			result = false;
		}
		return result;
	}

	void __log_err_inv_prop(std::string v)
	{
		boost::format fmt = boost::format("Unknown Config property: %1%") % v;
		LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
	}

	double ConfigHandler::getDouble(std::string v)
	{
		std::map<std::string, std::string>::iterator iValue = _confValues.find(v);
		if(iValue != _confValues.end())
		{
			return std::stod(iValue->second);
		}
		else
		{
			__log_err_inv_prop(v);
		}
		return 0;
	}

	ui16 ConfigHandler::getUInt16(std::string v)
	{
		std::map<std::string, std::string>::iterator iValue = _confValues.find(v);
		if(iValue != _confValues.end())
		{
			return static_cast<ui16>(std::stoi(iValue->second));
		}
		else
		{
			__log_err_inv_prop(v);
		}
		return 0;
	}

	ui32 ConfigHandler::getUInt32(std::string v)
	{
		std::map<std::string, std::string>::iterator iValue = _confValues.find(v);
		if(iValue != _confValues.end())
		{
			return std::stoi(iValue->second);
		}
		else
		{
			__log_err_inv_prop(v);
		}
		return 0;
	}

	ui64 ConfigHandler::getUInt64(std::string v)
	{
		std::map<std::string, std::string>::iterator iValue = _confValues.find(v);
		if(iValue != _confValues.end())
		{
			return std::stoull(iValue->second);
		}
		else
		{
			__log_err_inv_prop(v);
		}
		return 0;
	}

	std::string ConfigHandler::getPath(std::string v)
	{
		std::map<std::string, std::string>::iterator iValue = _confValues.find(v);
		if(iValue != _confValues.end())
		{
			return iValue->second;
		}
		else
		{
			__log_err_inv_prop(v);
		}
		return "";
	}
}