/*
Serial traffic simulation program for PaPo.

############################################

This application uses data from Open Street Map, henceforth
that data is made available under the Open Database License: http://opendatacommons.org/licenses/odbl/1.0/.
Any rights in individual contents ofthe database are licensed under 
the Database Contents License: http://opendatacommons.org/licenses/dbcl/1.0/
See more at: http://opendatacommons.org/licenses/odbl/#sthash.y0L9Yy1N.dpuf

############################################

Author: Martin Poppinga, Miriam von Heeremann, Marcel Ellermann
Date: 24.07.2013      
Distributed under the Boost Software License, Version 1.0.
See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)

#############################################
*/

#include "Log.h"
#include "boost/date_time/posix_time/posix_time.hpp"

namespace traffic_simulation
{
	bool Log::_bInstance = false;
	Log* Log::_instance = NULL;

	/// <summary>
	/// Returns the instance of our <see cref="Log" /> class to secure that only a single instance is created.
	/// </summary>
	/// <returns></returns>
	Log* Log::getInstance()
	{
		if(!_bInstance)
		{
			_instance = new Log();
			_bInstance = true;
		}
		return _instance;
	}

	/// <summary>
	/// Prevents a default instance of the <see cref="Log"/> class from being created.
	/// </summary>
	Log::Log()
	{
		_streetLog = new std::ofstream(STREET_LOG_FILE, std::ofstream::binary);
		_carLog = new std::ofstream(CAR_LOG_FILE, std::ofstream::binary);
		_crossingLog = new std::ofstream(CROSSING_LOG_FILE, std::ofstream::binary);
		_worldLog = new std::ofstream(WORLD_LOG_FILE, std::ofstream::binary);
		_trafficLog = new std::ofstream(TRAFFIC_LOG_FILE, std::ofstream::binary);
	}

	Log::~Log()
	{
		//free allocated memory, closing of stream is done within the ofstream deconstructor
		delete _streetLog;
		delete _carLog;
		delete _crossingLog;
		delete _worldLog;
		delete _trafficLog;
		_bInstance = false;
		_instance = NULL;
	}

	/// <summary>
	/// Writes the current time stamp to the output stream.
	/// </summary>
	/// <param name="s">The stream to write the time stamp to.</param>
	void Log::writeTimeStamp(std::ofstream* s, boost::uint32_t timeStamp)
	{		
		//get local system time
		boost::posix_time::ptime cur(boost::posix_time::second_clock::local_time());
		std::string sout = boost::posix_time::to_simple_string(cur);
		(*s) << "[" << sout << "+" << timeStamp << ":" << world::getCommRank() << "] ";
	}

	void Log::writeLogOut(std::ofstream* out, const char* logText, boost::uint32_t timeStamp)
	{
		writeTimeStamp(out, timeStamp);
		(*out) << logText << std::endl;
		out->flush();
	}

	/// <summary>
	/// Writes the street log.
	/// </summary>
	/// <param name="logText">The log text.</param>
	void Log::writeStreetLog(const char* logText)
	{	
		writeLogOut(_streetLog, logText, WORLD_TIME_STAMP);
#ifdef LOG_PRINT_STREET_LOG
		std::cout << "STREET: " << logText << std::endl;
#endif
	}

	/// <summary>
	/// Writes the car log.
	/// </summary>
	/// <param name="logText">The log text.</param>
	void Log::writeCarLog(const char* logText)
	{
		writeLogOut(_carLog, logText,WORLD_TIME_STAMP);
#ifdef LOG_PRINT_CAR_LOG
		std::cout << "CAR: " << logText << std::endl;
#endif
	}

	/// <summary>
	/// Writes the traffic log.
	/// </summary>
	/// <param name="logText">The log text.</param>
	void Log::writeTrafficLog(const char* logText)
	{
		writeLogOut(_trafficLog, logText, WORLD_TIME_STAMP);
#ifdef LOG_PRINT_TRAFFIC_LOG
		std::cout << "TRAFFIC: " << logText << std::endl;
#endif
	}

	/// <summary>
	/// Writes the world log.
	/// </summary>
	/// <param name="logText">The log text.</param>
	void Log::writeWorldLog(const char* logText)
	{
		writeLogOut(_worldLog, logText, WORLD_TIME_STAMP);
#ifdef LOG_PRINT_WORLD_LOG
		std::cout << "WORLD: " << logText << std::endl;
#endif
	}
}