/*
Serial traffic simulation program for PaPo.

############################################

This application uses data from Open Street Map, henceforth
that data is made available under the Open Database License: http://opendatacommons.org/licenses/odbl/1.0/.
Any rights in individual contents ofthe database are licensed under 
the Database Contents License: http://opendatacommons.org/licenses/dbcl/1.0/
See more at: http://opendatacommons.org/licenses/odbl/#sthash.y0L9Yy1N.dpuf

############################################

Author: Martin Poppinga, Miriam von Heeremann, Marcel Ellermann
Date: 24.07.2013      
Distributed under the Boost Software License, Version 1.0.
See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)

#############################################
*/

#include "boost/cstdint.hpp"
#include "world.h"
#include "worldDefs.h"
#include "MapHandler.h"
#include "SpawnHandler.h"
#include "Log.h"
#include "ConfigHandler.h"
#include "boost/format.hpp"
#include "mpi.h"
#include "Messages.h"
#include "MessageHandler.h"
#include <time.h>

using namespace boost;
using namespace traffic_simulation;


// Main Method to initialize the programm
// may pass args like how many cars should be generated and so on?
int main(int argc, char* argv[])
{

#ifdef _OPENMP
	double elapsedTime = omp_get_wtime( );
	omp_set_nested(1);
#else
	clock_t elapsedTime = clock();
#endif
	int available = 0;

	MPI_Init_thread(&argc, &argv, MPI_THREAD_MULTIPLE, &available);
	if(available < MPI_THREAD_SERIALIZED)
	{
		LOG->writeWorldLog("Unable to start program. MPI_THREAD_SERIALIZED not supported.");
	}
	else
	{
		MessageHandler::Init(available);
#pragma omp parallel
		{
#pragma omp sections
			{
#pragma omp section
				{
					MPIRcvMessage_CtoB _CtoBMsg;
					MPIRcvMessage_Car _carMsg;
					MPIRcvMessage_CtoE _CtoEMsg;
					MPIRcvMessage_NCtoB _NCtoB;
					MPIRcvMessage_NCtoE _NCtoE;
					MessageHandler::getInstance()->RunRcvManager();
				}
#pragma omp section
				{
					world* w = world::getInstance();
					boost::format mpiTSupport = boost::format("Starting Simulation with MPI Threadsupport: %1% and %2% threads.") % available % world::getCommSize();
					LOG->writeWorldLog(_FMT_TO_C_STR_(mpiTSupport));
					if(world::loaded())

					{
						ui32 i = 0;
						ui32 max = SIMULATION_TICKS;
						while(i < max)
						{
							w->update();
							MSG->WaitMsgsSend();
							MSG->Synchronize();
							i++;
						}
						MSG->FinalizeHandler();
					}
#ifdef _OPENMP
					elapsedTime = omp_get_wtime( ) - elapsedTime;
					boost::format fmt = boost::format("Executed %1% updates in %2% seconds") % SIMULATION_TICKS % (elapsedTime);
#else
					elapsedTime = clock() - elapsedTime;
					boost::format fmt = boost::format("Executed %1% updates in %2% seconds") % SIMULATION_TICKS % (((float)elapsedTime)/CLOCKS_PER_SEC);
#endif
					LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
				}
			}
		}
	}
	MPI_Finalize();
	return 0;
}
