/*
Serial traffic simulation program for PaPo.

############################################

This application uses data from Open Street Map, henceforth
that data is made available under the Open Database License: http://opendatacommons.org/licenses/odbl/1.0/.
Any rights in individual contents ofthe database are licensed under
the Database Contents License: http://opendatacommons.org/licenses/dbcl/1.0/
See more at: http://opendatacommons.org/licenses/odbl/#sthash.y0L9Yy1N.dpuf

############################################

Author: Martin Poppinga, Miriam von Heeremann, Marcel Ellermann
Date: 24.07.2013
Distributed under the Boost Software License, Version 1.0.
See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)

#############################################
*/

#include "MapHandler.h"
#include "boost/format.hpp"
#include "boost/heap/binomial_heap.hpp"
#include "SpawnHandler.h"
#include "ConfigHandler.h"
#include "Log.h"
#include <omp.h>
#include <list>
#include <sstream>
#include "MessageHandler.h"
#include "Messages.h"

using namespace std;

namespace traffic_simulation
{
	bool MapHandler::_bInstance = false;
	bool MapHandler::_bLoaded = false;
	MapHandler* MapHandler::_instance = NULL;

	MapHandler::MapHandler()
	{
		if(NodeHandler::getInstance()->loaded())
		{
			bool rsl = loadStreets();
			bool ril = loadIntersections();

			_bLoaded = (rsl && ril);

			//check intersection points
		}
		else
		{
			LOG->writeWorldLog("Cannot initialize MapHandler, NodeHandler did not load successfully.");
		}
	}

	MapHandler::~MapHandler()
	{
		_instance = NULL;
		_bInstance = false;
		_bLoaded = false;

		//free streets
		for(std::vector<std::vector<Street*>>::iterator iSList = _streets.begin(); iSList != _streets.end(); iSList++)
		{
			for(std::vector<Street*>::iterator itr = iSList->begin(); itr != iSList->end(); itr++)
			{
				delete(*itr);
			}
			iSList->clear();
		}
		_streets.clear();
	}

	MapHandler* MapHandler::getInstance()
	{
		if(!_bInstance)
		{
			_instance = new MapHandler();
			_bInstance = true;
		}
		return _instance;
	}

	/// <summary>
	/// Gets the bucket that contains the given street.
	/// Also transforms the ID into the buckets internal storage ID
	/// Therefore we can access the street directly by the returned value and the sID
	/// </summary>
	/// <param name="sID">The street identifier.</param>
	/// <returns>The ID of the Bucket the given street is stored in</returns>
	int MapHandler::getBucket(ui32 &sID)
	{
		int result = -1;
		if(sID >= 0 && sID < _streetCount)
		{
			int c = 0;
			int size = (int)_streets[c].size();
			boost::int64_t id = static_cast<boost::int64_t>(sID);
			while(id >  size - 1)
			{
				sID -= _streets[c].size();
				c++;
				size += (int)_streets[c].size();
			}
			result = c;
		}
		return result;
	}
	bool is_perfect_square(int n) {
		if (n < 0)
			return false;
		int a = (int) sqrt((double) n);
		return (a * a == n || (a+1)*(a+1) == n);
	}
	/// <summary>
	/// Sorts the streets into the number of buckets defined by MPI_COMM_SIZE
	/// </summary>
	void MapHandler::sortStreets()
	{
		if(!is_perfect_square(world::getCommSize()))
		{
			LOG->writeWorldLog("Invalid number of processes, number cannot be uneven.");
		}
		else
		{
			/*
			Only accept squres to form a good raster above the map
			2x2 or 4x4
			*/
			int sNum = (int)sqrt((double)world::getCommSize()); //number of raster per row / column
			double deltaX = NODES->getMax()->getX() - NODES->getMin()->getX();
			double deltaY = NODES->getMax()->getY() - NODES->getMin()->getY();
			//Map graph has a width of deltaX and a height of deltaY
			double rasterWidth = deltaX / sNum; //widht of one raster
			double rasterHeight = deltaY / sNum; //height of one raster

			//iterate over every street and assign them to a bucket that represents its' raster
			for(std::vector<Street*>::iterator iStreet = _streets[world::getCommSize()].begin(); iStreet != _streets[world::getCommSize()].end(); iStreet++)
			{
				//determin a streets bucket by its starting coordinates
				const worldPos* p = (*(*iStreet)->nodes()->begin())->getPos();
				double diffX = p->getX() - NODES->getMin()->getX();
				double diffY = p->getY() - NODES->getMin()->getY();
				int rX = (int)(diffX / rasterWidth);
				int rY = (int)(diffY / rasterHeight);
				//now we got our position on the raster, now compute a lineazed number
				// 0 1 2   0,0 1,0 2,0   therfore 0,0 => 0 
				// 3 4 5   1,0 1,1 2,1
				// 6 7 8   2,0 2,1 2,2
				//index within the storage vector
				int vIdx = rY * sNum + rX;
				//store the bucket id in every street
				(*iStreet)->setBucket(vIdx);
				_streets[vIdx].push_back(*iStreet);
			}
			_streets[world::getCommSize()].clear(); // sorted all streets, temporary storage

			//can be removed, be sure to not store pointers to those locations, they'll be invalid
			//set new id's for the streets to make access to'em easier later
			//iterator over every bucket
			ui32 idCounter = 0;
#ifdef CAR_DUMP
			boost::format fmt = boost::format("%1%_mapping.dat") % PATH_CAR_DUMP; 
			std::ofstream streetMapFile(_FMT_TO_C_STR_(fmt), std::ios::binary);
#endif
			for(std::vector<std::vector<Street*>>::iterator iSVector = _streets.begin(); iSVector != _streets.end(); iSVector++)
			{
				//iterator over every street within the bucket
				for(std::vector<Street*>::iterator iStreet = iSVector->begin(); iStreet != iSVector->end(); iStreet++)
				{
					ui32 oldID = (*iStreet)->id();
#ifdef CAR_DUMP
					streetMapFile.write((char*)(&oldID), sizeof(oldID));
					streetMapFile.write((char*)(&idCounter), sizeof(idCounter));
#endif
					(*iStreet)->setID(idCounter);
					idCounter++;
				}
			}
#ifdef CAR_DUMP
			streetMapFile.close();
#endif
		}
	}

	/// <summary>
	/// Get street from Street index.
	/// </summary>
	/// <param name="id">The id of the street.</param>
	/// <returns>Pointer to the street with the given id. NULL if street could not be found.	</returns>
	Street* MapHandler::getStreet(ui32 id)
	{
		Street* result = NULL;
		int c = getBucket(id);
		if(c > -1)
		{
			result =  _streets[c][id];
		}
		else
		{
			boost::format fmt = boost::format("Street ID out of bounds. Invalid Id: %1%") % id;
			LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
		}
		return result;
	}

	/// <summary>
	/// Loads the intersections.
	/// </summary>
	/// <returns>True if no errors occured while loading the intersections</returns>
	bool MapHandler::loadIntersections()
	{
		bool success = true;
		stringstream inputFile;
		inputFile << world::getInputFile(FILE_INTERSECTIONS);
		LOG->writeWorldLog("Start loading Intersections...");
		try
		{
			//open file
			ifstream in(inputFile.str());
			//check if file was opened
			if(in)
			{
				std::string Version;
				std::getline(in,Version);
				if(CHECK_VERSION(Version.c_str()))
				{
					//go through every line
					//structure: (uInt)id,(b)trafficLight,(uInt)id
					while(!in.eof() && in.peek() > 30 /*only writable character*/)
					{
						ui32 nodeID = 0;
						bool tl = false;
						string sID1, sIDs, sTL,sSpawn, sPWays;
						getline(in, sID1, ';'); //read line till delimiter
						getline(in, sTL, ';');
						getline(in, sSpawn, ';');
						getline(in, sIDs, ';');
						getline(in, sPWays); //currently unused
						//convert strings to actual datatypes
						//won't use an extra try catch block here since we're already in one
						nodeID = stoul(sID1);
						tl  = (sTL == "True");
						//load ways from sIDs
						stringstream ss(sIDs);
						string way;

						list<Street*> streets;

						while(getline(ss, way, ':'))
						{
							ui32 id = stoul(way); //convert string to street id
							Street* s = getStreet(id);
							if(s) //get street from store
							{
								streets.push_back(s);
							}
							else
							{
								stringstream sOut;
								sOut << "Could not find Street: " << id;
								LOG->writeWorldLog(sOut.str().c_str());
								success = false;
							}
						}
						map<ui32, map<ui32,list<Street*>>> priorityWays;
						//load priorityWays from input
						ui32 srcWay,destWay,priWays;
						string priorityRule;
						stringstream rules(sPWays);
						while(getline(rules, priorityRule, ']')) // ] is delimiter for rules
						{
							stringstream ruleReader(priorityRule);
							//extract srcWay
							ruleReader >> srcWay;
							ruleReader.ignore(1,'['); //ignore open bracket for destination ways
							//read all target ways
							string destRule;
							while(getline(ruleReader, destRule, ';'))
							{
								stringstream destRuleReader(destRule);
								destRuleReader >> destWay;
								destRuleReader.ignore(1, ','); //ignore destWay delimiter
								string sDestWay;
								list<Street*> LpriorityWays;
								while(getline(destRuleReader, sDestWay, ','))
								{
									priWays = stoul(sDestWay);
									Street* priStreet = getStreet(priWays);
									if(priStreet)
									{
										LpriorityWays.push_back(priStreet);
									}
									else
									{
										stringstream logoutput;
										logoutput << "Couldn't find Street: " << priWays;
										LOG->writeWorldLog(logoutput.str().c_str());
									}
								}
								priorityWays[srcWay][destWay] = LpriorityWays;
							}
						}
						//all streets loaded from line
						//add intersections to respective streets now
						//only if we have at least 2 streets connected in our intersection, load it -- spawner and others currently unhandled
						if(streets.size() > 1)
						{
							//loaded data from line, now check for correctness
							Node* n1 =  NodeHandler::getInstance()->getNode(nodeID);
							if((n1))
							{
								//add intersections to streets
								for(list<Street*>::iterator s1 = streets.begin(); s1 != streets.end(); s1++)
								{
									//clone iterator here
									list<Street*>::iterator s2 = list<Street*>::iterator(s1);
									/*
									Were iterating over the list now with the second iterator and start from the beginning when we reach
									the end until we reach s2 again. This way we keep the order of streets for every respectiv street.
									*/
									s2++;
									if(s2 == streets.end())
									{
										s2 = streets.begin();
									}
									while(s2 != s1)
									{
										//pointer to our actual streets
										Street *street1, *street2;
										street1 = (*s1);
										street2 = (*s2);
										bool addS2 = true; //add intrersection from street1 to street2, to street1
										if(street1->isOneWay())
										{
											//if one way is in forward direction, and our intersection is at the start node, we can't get
											//from street1 to street 2
											if(street1->oneWayDirection() == driveDirection::forward && (*street1->nodes()->begin())->getID() == nodeID)
											{
												addS2 = false;
											}
											//the same just other whise round
											if(street1->oneWayDirection() == driveDirection::backward && (*street1->nodes()->rbegin())->getID() == nodeID)
											{
												addS2 = false;
											}
										}
										if(street2->isOneWay())
										{
											if(street2->oneWayDirection() == driveDirection::forward  && (*street2->nodes()->rbegin())->getID() == nodeID)
											{
												addS2 = false;
											}
											if(street2->oneWayDirection() == driveDirection::backward && (*street2->nodes()->begin())->getID() == nodeID)
											{
												addS2 = false;
											}
										}
										if(addS2) // intersection
										{
											street1->addIntersection(nodeID, street2, priorityWays[street1->id()][street2->id()]);
										}
										s2++; //increase our iterator.
										if(s2 == streets.end())
										{
											s2 = streets.begin();
										}
									}
								}
							}
							else
							{
								stringstream sOut;
								sOut << "Invalid node required for intersection. Node: " << nodeID;
								LOG->writeWorldLog(sOut.str().c_str());
							}
						}
					}
				}
				else
				{
					boost::format fmt = boost::format("Invalid File Version: %1%. Only Version: %2% supported.") % Version % MAP_DATA_VERSION;
					LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
				}
			}
			else
			{
				success = false;
				stringstream sOut;
				sOut << "Could not open Intersection File: " << inputFile.str();
				LOG->writeWorldLog(sOut.str().c_str());
			}
		}
		catch(std::exception e)
		{
			success = false;
			LOG->writeWorldLog("Error while loading Intersections.");
			LOG->writeWorldLog(e.what());
		}
		LOG->writeWorldLog("finished Loading intersections.");
		return success;
	}

	/// <summary>
	/// Gets the streets at node n.
	/// </summary>
	/// <param name="n">The node.</param>
	/// <returns>A list of streets that may intersect at node n</returns>
	std::list<Street*> MapHandler::getStreetsAtNode(Node* n)
	{
		std::list<Street*> rList;
		for(std::vector<std::vector<Street*>>::iterator iSVector = _streets.begin(); iSVector != _streets.end(); iSVector++)
		{
			for(std::vector<Street*>::iterator itr = iSVector->begin(); itr != iSVector->end(); itr++)
			{
				for(std::list<Node*>::const_iterator iNode = (*itr)->nodes()->begin(); (*itr)->nodes()->end() != iNode; iNode++)
				{
					if(*iNode == n)
					{
						rList.push_back(*itr);
						break;
					}
				}
			}
		}
		return rList;
	}

	void MapHandler::changeStreet(Car*  c, ui32 oldStreet, ui32 gNodeID, ui32 gStreetID)
	{
		//handle movements from cars to other streets here
		getStreet(oldStreet)->setCarRemove();
		Street* goalStreet = getStreet(gStreetID);
		if(goalStreet->getBucket() == world::getCommRank())
		{
			//add car to our new street
			goalStreet->addCar(c, gNodeID);
		}
		else
		{
			__carStoreStruct _mCar;
			_mCar._cMePosition = c->getMetricPos();
			_mCar._cSpeed = c->getCurrentSpeed();
			_mCar._cStreet = gStreetID;
			_mCar._guid = c->getGUID();
			int i = 0;
			for(std::list<std::pair<Node*,Street*>*>::const_iterator _pItr = c->getPath()->begin(); _pItr != c->getPath()->end(); _pItr++)
			{
				_mCar._path[i][0] = (*_pItr)->first->getID();
				_mCar._path[i][1] = (*_pItr)->second->id();
				i++;
				if(i == 700)
					break;
			}
			_mCar.pathSize = i;
			MSG->SendMessage(&_mCar, 1, MSG_CAR, goalStreet->getBucket());
			SPAWNS->registerForRemove(c);
		}
	}

	void MapHandler::update(ui8 step)
	{
		int i;
		ui32 max = _streets[world::getCommRank()].size();
#pragma omp parallel for private(i) shared(max) schedule(guided)
		for(i = 0; (ui32)i < max; i++)
		{
			Street* s = _streets[world::getCommRank()][i];
			s->update(step);
		}
	}

	bool MapHandler::loadStreets()
	{
		bool success = true;
		stringstream inputFile;
		inputFile << world::getInputFile(FILE_STREETS);
		LOG->writeWorldLog("Start loading Streets...");
		try
		{
			ifstream in(inputFile.str());
			if(in)
			{
				std::string Version;
				std::getline(in,Version);
				if(CHECK_VERSION(Version.c_str()))
				{
					std::string c_Streets;
					std::getline(in,c_Streets);
					_streetCount = std::stoul(c_Streets);
					_streets.resize(world::getCommSize() + 1);
					_streets[world::getCommSize()].resize(_streetCount);
					std::vector<Street*>* sVector = &_streets[world::getCommSize()];
					while(!in.eof() && in.peek() > 30 /*only writable characters*/)
					{
						Street* s = new Street();
						in >> (*s);
						if((*sVector)[s->id()] != NULL)
						{
							stringstream dOut;
							dOut << "Error while loading streets, duplicate entry for street: " << s->id();
							LOG->writeWorldLog(dOut.str().c_str());
							success = false;
						}
						else
						{
							(*sVector)[s->id()] = s;
						}
					}
					LOG->writeWorldLog("Finished Loading Streets");
				}
				else
				{
					boost::format fmt = boost::format("Invalid File Version: %1%. Only Version: %2% supported.") % Version % MAP_DATA_VERSION;
					LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
				}
			}
			else
			{
				stringstream eOut;
				eOut << "Unable to find street data file at: " << inputFile.str();
				LOG->writeWorldLog(eOut.str().c_str());
				success = false;
			}
		}
		catch(std::exception e)
		{
			LOG->writeWorldLog("Error while loading the streets.");
			LOG->writeWorldLog(e.what());
			success = false;
		}
		return success;
	}

	bool operator > (MapHandler::openListNode left, MapHandler::openListNode right)
	{
		return left.get<0>() > right.get<0>();
	}

	template<typename G, typename H>
	std::list<std::pair<Node*, Street*>*>* MapHandler::getPath(std::pair<Node*,Street*> start, std::pair<Node*,Street*> end)
	{
		//open Nodes
		MapHandler::nodeHeap openList;
		//closed Nodes
		MapHandler::closedMap closedList;
		//stores handles to respectiv node / street pairs on the heap
		boost::unordered_map<pair<Node*,Street*>,nodeHeap::handle_type,__clHash> handleMap;
		//return path
		std::list<std::pair<Node*, Street*>*>* path = new list<std::pair<Node*, Street*>*>(); 
		//value tuple representing our current node in the graph
		openListNode* currentNode = NULL;

		bool foundPath = false;
		//push start node to openList and store handle
		handleMap[start] = openList.push(new openListNode(0, 0, start, NULL));

		while(!openList.empty())
		{
			//get min node from heap
			currentNode = openList.top();

			//remove top element from heap
			openList.pop();
			handleMap.erase(handleMap.find(currentNode->get<2>())); //erase handle from handle map - it's not valid anymore
			//add currentNode to closed list
			closedList[currentNode->get<2>()] = currentNode;
			Street* curStreet = currentNode->get<2>().second;
			Node* curNode = currentNode->get<2>().first;
			//add new nodes to openlist
			//start with successiv street nodes
			list<Node*>::const_iterator nItr = curStreet->nodes()->begin();
			if(curStreet->id() == end.second->id())
			{
				openListNode* x = currentNode;
				currentNode = NULL;
				if(end.first->getID() != curNode->getID()) // end node is not the current node
					path->push_back(new pair<Node*, Street*>(end.first,curStreet));
				while(x->get<2>().first->getID() != start.first->getID() || x->get<2>().second->id() != start.second->id())
				{
					path->push_front(new pair<Node*, Street*>(x->get<2>()));
					openListNode* __tmp = x;
					x = (openListNode*)x->get<3>();
					closedList.erase(closedList.find(__tmp->get<2>()));
					delete(__tmp);
				}
				closedList.erase(closedList.find(x->get<2>()));
				delete x;
				foundPath = true;
				break;
			}
			if(foundPath)
				break;
			for(std::multimap<ui32,std::pair<Street*,std::list<Street*>>>::const_iterator itr = curStreet->intersections()->begin();
				itr != curStreet->intersections()->end(); itr++)
			{
				pair<Node*,Street*> p(NodeHandler::getInstance()->getNode(itr->first),itr->second.first);
				if(closedList.find(p) == closedList.end())
				{
					G g; //cost function
					H h; //heuristic function
					double cost = g(currentNode, &p);
					unordered_map<pair<Node*,Street*>,nodeHeap::handle_type>::iterator t;
					if((t = handleMap.find(p)) != handleMap.end())
					{

						if((*t->second)->get<1>() > cost)
						{
							(*t->second)->get<0>() = cost + h(&p, &end);
							(*t->second)->get<1>() = cost; //update costs
							(*t->second)->get<3>() = currentNode; //set new src node
							openList.increase(t->second);
						}
					}
					else
					{
						nodeHeap::handle_type t = openList.push(new openListNode(cost + h(&p, &end), cost, p, currentNode));
						handleMap[p] = t;
					}
				}
			}
		}
		//cleanup
		unordered_map<pair<Node*,Street*>,nodeHeap::handle_type,__clHash>::iterator itr = handleMap.begin();
		while(itr != handleMap.end())
		{
			openListNode* node = (*itr->second);
			delete(node);
			itr++;
		}
		//cleanup allocated memeory for closed list
		for(MapHandler::closedMap::iterator iCMap = closedList.begin(); iCMap != closedList.end();)
		{
			delete(iCMap->second);
			iCMap++;
		}
		return path;

	}
	//add other possible path cost calculation templates here
	template std::list<std::pair<Node*, Street*>*>* MapHandler::getPath<__simpleCost, __euclHeuristic>(std::pair<Node*,Street*> start, std::pair<Node*,Street*> end);
	template std::list<std::pair<Node*, Street*>*>* MapHandler::getPath<__speedCost, __euclHeuristic>(std::pair<Node*,Street*> start, std::pair<Node*,Street*> end);

}
