/*
Serial traffic simulation program for PaPo.

############################################

This application uses data from Open Street Map, henceforth
that data is made available under the Open Database License: http://opendatacommons.org/licenses/odbl/1.0/.
Any rights in individual contents ofthe database are licensed under 
the Database Contents License: http://opendatacommons.org/licenses/dbcl/1.0/
See more at: http://opendatacommons.org/licenses/odbl/#sthash.y0L9Yy1N.dpuf

############################################

Author: Martin Poppinga, Miriam von Heeremann, Marcel Ellermann
Date: 24.07.2013      
Distributed under the Boost Software License, Version 1.0.
See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)

#############################################
*/
#ifndef H_MAP_HANDLER
#define H_MAP_HANDLER

#include <fstream>
#include "Street.h"
#include "WorldObj.h"
#include "CarHandler.h"
#include "worldDefs.h"
#include "boost/heap/binomial_heap.hpp"


#pragma once

namespace traffic_simulation
{
	struct __clHash;
	template<class _Ty>	struct oLngreater;

	class MapHandler
	{
	public:
		typedef boost::tuple<double, double, std::pair<Node*, Street*>, void* /*ptr to openListNode*/> openListNode;
		typedef boost::heap::binomial_heap<openListNode*,boost::heap::compare<oLngreater<openListNode*>>> nodeHeap;
		typedef boost::unordered_map<std::pair<Node*, Street*>,openListNode*,__clHash> closedMap;
		static MapHandler* getInstance();
		static bool loaded(){return _bLoaded;};
		CarHandler* getCarHandler(){return &_carhandler;};
		int getBucket(ui32 &sID);

		//generates a path from a start node to a goal node, since goal and starts usually are spawners we can use plain nodes here
		template<typename G, typename H>
		std::list<std::pair<Node*, Street*>*>* getPath(std::pair<Node*,Street*> start, std::pair<Node*,Street*> end);
		ui32 streetCount(){return _streetCount;}; //returns number of streets
		std::list<Street*> getStreetsAtNode(Node* n);
		void changeStreet(Car* c, ui32 oldStreet, ui32 gNodeID, ui32 gStreetID);
		void sortStreets();
		Street* getStreet(ui32 id);
		void update(ui8 step = 1);
		~MapHandler();
	private:
		MapHandler();
		bool loadStreets();
		bool loadIntersections();
		std::vector<std::vector<Street*>> _streets;
		CarHandler _carhandler;
		static MapHandler* _instance;
		static bool _bInstance;
		static bool _bLoaded;
		ui32 _streetCount;
	};

	struct __clHash : std::unary_function<std::pair<Node*, Street*>, std::size_t>
	{
		std::size_t operator()(std::pair<Node*,Street*> const& val) const
		{
			return val.second->id();
		}
	};


	struct __euclHeuristic
	{
		double operator()(std::pair<Node*, Street*> *destNode, std::pair<Node*, Street*> *endNode)
		{
			return destNode->first->getPos()->distance(endNode->first->getPos());
		}
	};

	struct __simpleCost
	{
		double operator()(MapHandler::openListNode *olNode, std::pair<Node*, Street*>* destNode)
		{
			return olNode->get<1>() + destNode->second->length();
		}
	};

	struct __speedCost
	{
		double operator()(MapHandler::openListNode* olNode, std::pair<Node*, Street*>* destNode)
		{
			//drive direction doesn't matter becaus we just want the difference
			driveDirection d = driveDirection::forward;
			double dist = olNode->get<2>().second->getNodePosition(olNode->get<2>().first, d) - olNode->get<2>().second->getNodePosition(destNode->first, d);
			return olNode->get<1>()/*old costs*/ + fabs(dist)/ (destNode->second->speed());
		}
	};

		/*
	Overloads the > operator for our tuples to allow > comparision
	*/
	extern bool operator > (MapHandler::openListNode left, MapHandler::openListNode right);

	template<class _Ty>
	struct oLngreater
		: public std::binary_function<_Ty, _Ty, bool>
	{	// functor for operator>
	bool operator()(const _Ty& _Left, const _Ty& _Right) const
		{	// apply operator> to operands
		return ((*_Left) > (*_Right));
		}
	};
}
#endif