#include "MessageHandler.h"
#include "Messages.h"
#include <stdint.h>
#include <vector>


MessageHandler* MessageHandler::_instance = NULL;

void IMPIRcvMessage::registerSelf()
{
	if(MessageHandler::IsRunning())
	{
#pragma omp critical(MsgHandlerAccess)
		{
			MessageHandler::getInstance()->_msgHandler.insert(std::make_pair(_msgID, this));
		}
	}
	else
	{
#pragma omp critical(cout)
		{
			std::cout << "Could register Message add Message handler. Message handler not running." << std::endl;
		}
	}
}

/// <summary>
/// Initializes the an instance of the MessageHandler.
/// The instance can be accessed with the provided
/// getInstance() method. Make sure to call this Method
/// before u fork into multiple thread to assure that 
/// a valid instance of the MessageHandler was created.
/// </summary>
/// <param name="mpiThreadSupport">The mpi thread support.</param>
void MessageHandler::Init(int mpiThreadSupport)
{
	if(_instance == NULL)
	{
		_instance = new MessageHandler();
		_instance->_isRunning = false;
		_instance->_stopAcceptMessage = false;
		_instance->_receiveMessages = true;
		_instance->_finalizeHandler = false;
		_instance->_readyForRcv = false;
		_instance->_mpiThreadSupport = mpiThreadSupport;
#pragma omp critical (MPI_FUNC_CALL)
		{
			MPI_Comm_rank(MPI_COMM_WORLD,&_instance->_rank);
			MPI_Comm_size(MPI_COMM_WORLD,&_instance->_commSize);
		}
	}
}

MessageHandler* MessageHandler::getInstance()
{
	return _instance;
}

/// <summary>
/// Starts the rcv handler. The Receive handler receives messages, that have regestired itself
/// at the message handler. This method is blocking, so it is recommend to star it within a own thread.
/// call MessageHandler::Finalize() to  stop  the receive handler
/// </summary>
void MessageHandler::RunRcvManager()
{
	_message_rcv_handler();
}

void MessageHandler::SendMessage(void* val, int count, int msgCode, int rcvr)
{
	if(!_stopAcceptMessage)
	{
		std::map<int, IMPIRcvMessage*>::iterator iMsgMap = _msgHandler.find(msgCode);
		if(iMsgMap != _msgHandler.end())
		{
			MPI_Aint lb, size;
			MPI_Type_get_true_extent(iMsgMap->second->getDataType(), &lb, &size);
			MPIMsgContainer* msg = new MPIMsgContainer;
			msg->val = malloc(iMsgMap->second->getValueSize());
			memcpy(msg->val, val, iMsgMap->second->getValueSize());
			msg->type = iMsgMap->second->getDataType();
			if(_mpiThreadSupport < 3)
			{
#pragma omp critical (MPI_FUNC_CALL)
				{
					MPI_Issend(msg->val, count, iMsgMap->second->getDataType(), rcvr, msgCode, MPI_COMM_WORLD, &msg->request);
				}
			}
			else
			{
				MPI_Issend(msg->val, count, iMsgMap->second->getDataType(), rcvr, msgCode, MPI_COMM_WORLD, &msg->request);
			}
#pragma omp critical (cout)
			{
				std::cout << "Send Message " << msgCode << " With Value: " << *(int*)msg->val << std::endl;
			}
#pragma omp critical (SendMsgListAccess)
			{
				_msgs[msg->request] = msg;
			}
		}
	}
	else
	{
#pragma omp critical(cout)
		{
			std::cout << "Unknown Message: " << msgCode << std::endl;
		}
	}
}

std::vector<int>* _getListenProcesses()
{
	std::vector<int>* r = new std::vector<int>();
	r->push_back(0);
	r->push_back(1);
	return r;
}


void MessageHandler::_message_rcv_handler()
{
	//init rcv listeners
	//get listerner count
	std::vector<int>* listeners = _getListenProcesses();
	//create array containing our request objects
	MPI_Request* msg_Requests = new MPI_Request[_msgHandler.size()];
	int* idxStore = new int[_msgHandler.size()];
	MPI_Status* rcvStatus = new MPI_Status[_msgHandler.size()];
	int c = 0;
	for(std::map<int, IMPIRcvMessage*>::iterator iRcv = _msgHandler.begin(); iRcv != _msgHandler.end(); iRcv++)
	{
#pragma omp critical (MPI_FUNC_CALL)
		{
			MPI_Irecv(iRcv->second->getValuePtr(), iRcv->second->getValueCount(), iRcv->second->getDataType(),
				iRcv->second->getSource(), iRcv->first, iRcv->second->getComm(), &msg_Requests[c]);
		}
		c++;
	}
	while(!_finalizeHandler)
	{
		int idx = -1;
		_readyForRcv = true;
		//ms mpi currently only availably for serialized threading
		//do critical section omp critical (MPI_FUNC_CALL) for every mpi calls
		int out = 0;
		int rcvErrStatus = MPI_SUCCESS;
		if(_mpiThreadSupport < 3)
		{
#pragma omp critical (MPI_FUNC_CALL)
			{
				rcvErrStatus = MPI_Testsome(_msgHandler.size(), msg_Requests, &out, idxStore, rcvStatus);
			}
		}
		else
			rcvErrStatus = MPI_Testsome(_msgHandler.size(), msg_Requests, &out, idxStore, rcvStatus);
		if(rcvErrStatus == MPI_SUCCESS)
		{
			if(out > 0)
			{
				for(int i = 0; i < out; i++)
				{
					int tag = rcvStatus[i].MPI_TAG;
					_msgHandler[tag]->handleMessage();
					if(_mpiThreadSupport < 3)
					{
#pragma omp critical (MPI_FUNC_CALL)
						{
							MPI_Irecv(_msgHandler[tag]->getValuePtr(), _msgHandler[tag]->getValueCount(), _msgHandler[tag]->getDataType(),
								_msgHandler[tag]->getSource(), tag, _msgHandler[tag]->getComm(), &msg_Requests[idxStore[i]]);
						}
					}
					else
					{
						MPI_Irecv(_msgHandler[tag]->getValuePtr(), _msgHandler[tag]->getValueCount(), _msgHandler[tag]->getDataType(),
							_msgHandler[tag]->getSource(), tag, _msgHandler[tag]->getComm(), &msg_Requests[idxStore[i]]);
					}
				}
			}
		}
		else
		{
#pragma omp critical (cout)
			{
				std::cout << rcvErrStatus << std::endl;
			}
		}
	}
}

/// <summary>
/// Synchronizes all MPI processes. Works like an MPI barrier
/// but this method is safe for the THREAD_SERIALIZED version
/// of MPI. Meaning that while the process is waiting for the
/// Barrier to be fullfilled, other MPI calls can be done by
/// another thread. If THREAD_MULTIPLE is available MPI_Barrier()
/// is the better option for synchronizing the threads.
/// </summary>
void MessageHandler::Synchronize()
{
	if(_mpiThreadSupport < 3)
	{
		/*sync_req[0]...sync_req[_commSize-1] : send Requests
		sync_req[commSize]...sync_req[_commSize*2-1] : recv Requests*/
		MPI_Request* sync_req = new MPI_Request[_commSize*2];
		/*sync_req[0]...sync_req[_commSize-1] : send Requests
		sync_req[commSize]...sync_req[_commSize*2-1] : recv Requests*/
		MPI_Status* sync_stat = new MPI_Status[_commSize*2];

		/*Send sync request to all processes within the Communicator
		And recieve sync messages from all process. Than wait till
		we got all Messages from all processes and all our messages
		haven been handled.*/
		int sendBuff = 1;
		int recvBuff = 1;
		for(int cRcvr = 0; cRcvr < _commSize; cRcvr++)
		{
#pragma omp critical (MPI_FUNC_CALL)
			{
				MPI_Issend(&sendBuff, 1, MPI_INTEGER, cRcvr, MSG_SYNC, MPI_COMM_WORLD, &sync_req[cRcvr]);
				MPI_Irecv(&recvBuff, 1, MPI_INTEGER, cRcvr, MSG_SYNC, MPI_COMM_WORLD, &sync_req[cRcvr + _commSize]);
			}
		}
		bool b_w_Sync = true;
		int flagHandled = 0;
		while(b_w_Sync)
		{
#pragma omp critical (MPI_FUNC_CALL)
			{
				MPI_Testall(_commSize * 2, sync_req, &flagHandled, sync_stat);
			}
			if(flagHandled == 1)
				b_w_Sync = false;
		}
	}
	else //The given MPI implementation supports MPI_THREAD_MULTIPLE 
	{
		//since we can call MPI function simultaniously by multiple threads, use a simple MPI_Barrier.
		MPI_Barrier(MPI_COMM_WORLD);
	}

}

void MessageHandler::WaitMsgsSend()
{ 
#pragma omp critical (SendMsgListAccess)
	{
		MPI_Request* request = new MPI_Request[_msgs.size()];
		MPI_Status* status = new MPI_Status[_msgs.size()];
		int c = 0;
		for(std::map<MPI_Request,MPIMsgContainer*>::iterator iMsg = _msgs.begin(); iMsg != _msgs.end(); iMsg++)
		{
			request[c] = iMsg->second->request;
			c++;
		}
		int sStatus = 0;
		int handled = 0;
		if(_mpiThreadSupport < 3)
		{
			while(handled == 0)
#pragma omp critical (MPI_FUNC_CALL)
			{
				MPI_Testall(_msgs.size(), request, &handled, status);
			}
		}
		else
		{
			MPI_Waitall(_msgs.size(), request, status);
		}
		for(std::map<MPI_Request,MPIMsgContainer*>::iterator iMsg = _msgs.begin(); iMsg != _msgs.end();)
		{
			free(iMsg->second->val);
			delete iMsg->second;
			iMsg = _msgs.erase(iMsg);
		}
	}
}

void MessageHandler::FinalizeHandler()
{
	_stopAcceptMessage = true;
	_finalizeHandler = true;
}
