/*
Serial traffic simulation program for PaPo.

############################################

This application uses data from Open Street Map, henceforth
that data is made available under the Open Database License: http://opendatacommons.org/licenses/odbl/1.0/.
Any rights in individual contents ofthe database are licensed under 
the Database Contents License: http://opendatacommons.org/licenses/dbcl/1.0/
See more at: http://opendatacommons.org/licenses/odbl/#sthash.y0L9Yy1N.dpuf

############################################

Author: Martin Poppinga, Miriam von Heeremann, Marcel Ellermann
Date: 24.07.2013      
Distributed under the Boost Software License, Version 1.0.
See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)

#############################################
*/

#include "NodeHandler.h"
#include "Car.h"
#include "Log.h"
#include "boost/format.hpp"
#include <omp.h>
#include <sstream>

namespace traffic_simulation
{
	bool NodeHandler::_bInstance = false;
	bool NodeHandler::_bLoaded = false;
	NodeHandler* NodeHandler::_instance = NULL;

	/// <summary>
	/// Returns the instance of our <see cref="NodeHandler" /> class to secure that only a single instance is created.
	/// </summary>
	/// <returns></returns>
	NodeHandler* NodeHandler::getInstance()
	{
		if(!_bInstance)
		{
			_instance = new NodeHandler();
			_bInstance = true;
		}
		return _instance;
	}

	/// <summary>
	/// Prevents a default instance of the <see cref="NodeHandler"/> class from being created.
	/// </summary>
	NodeHandler::NodeHandler()
	{
		if(loadNodes())
		{
			_bLoaded = true;
			_locks.resize(_nodes.size(), NULL);
#ifdef _OPENMP
			_lock_mutex.resize(_nodes.size(), NULL);
			for(unsigned i = 0; i < _nodes.size(); i++)
			{
				_lock_mutex[i] = new omp_lock_t;
				omp_init_lock(_lock_mutex[i]);
			}
#endif
		}
	}

	/// <summary>
	/// Finalizes an instance of the <see cref="NodeHandler"/> class.
	/// </summary>
	NodeHandler::~NodeHandler()
	{
		//remove all allocated Nodes
		while(!_nodes.empty())
		{
			std::vector<Node*>::iterator itr = _nodes.begin();
			delete *itr;
			_nodes.erase(itr);
		}
		_bInstance = false;
		_instance = NULL;
		_bLoaded = false;
		delete _min;
		delete _max;
	}
	std::ifstream& operator>>(std::ifstream& in, Node& d)
	{
		std::string sID, sX, sY;
		ui32 id;
		double x, y;

		std::getline(in, sID, ';');
		std::getline(in, sX, ';');
		std::getline(in, sY);

		try
		{
			id = std::stoul(sID);
			x = std::stod(sX);
			y = std::stod(sY);
			d.setID(id);
			d.getPos()->setPos(x,y);
		}
		catch(std::exception)
		{
			LOG->writeWorldLog("Couldn't convert node parameters from str.");
		}
		return in;
	}

	bool NodeHandler::loadNodes()
	{
		bool success = true;
		std::stringstream path;
		path << world::getInputFile(FILE_NODES);
		try
		{
#ifdef TF_DEBUG_MAIN
			LOG->writeWorldLog("Start loading Nodes from file...");
#endif
			std::ifstream in(path.str());
			std::string Version;
			std::getline(in,Version);
			if(CHECK_VERSION(Version.c_str()))
			{
				std::string nodeCount;
				std::getline(in, nodeCount);
				_nodes.resize(std::stoul(nodeCount));
				/*
					Extract Min and Max values for longitude & latitude
				*/
				std::string sMinMax;
				std::getline(in,sMinMax); 
				double mX,mY;
				std::stringstream streamMinMax(sMinMax);
				streamMinMax >> mX;
				streamMinMax.ignore(1, ';');
				streamMinMax >> mY;
				streamMinMax.ignore(1, ';');
				_min = new worldPos(mX,mY);
				streamMinMax >> mX;
				streamMinMax.ignore(1, ';');
				streamMinMax >> mY;
				_max = new worldPos(mX,mY);
				while(!in.eof() && in.peek() > 30 /*only writable characters*/)
				{
					Node* n = new Node(0,0,0);
					in >> (*n);
					if(_nodes[n->getID()] != NULL)
					{
#ifdef TF_DEBUG_MAIN
						std::string dEntry = "Duplicate entry found for Node";
						LOG->writeWorldLog(dEntry.c_str());
#endif
						success = false;
					}
					_nodes[n->getID()] = n;
				}
			}
			else
			{
				boost::format fmt = boost::format("Invalid File Version: %1%. Only Version: %2% supported.") % Version % MAP_DATA_VERSION;
				LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
			}
#ifdef TF_DEBUG_MAIN
			std::string finishNodes = "Finished loading Nodes, loaded Nodes: ";
			finishNodes += std::to_string((long double)_nodes.size());
			LOG->writeWorldLog(finishNodes.c_str());
#endif
		}
		catch(std::exception e)
		{
			LOG->writeWorldLog("Error while loading Node Data");
			LOG->writeWorldLog(e.what());
			success = false;
		}
		return success;
	}

	Node* NodeHandler::getNode(ui32 id)
	{
		Node* result = NULL;
		if(id < _nodes.size())
		{
			result = _nodes[id];
		}
		return result;
	}

	/// <summary>
	/// Adds a node lock for the current drive direction.
	/// </summary>
	/// <param name="id">The node identifier.</param>
	/// <param name="guid">The unique identifier.</param>
	/// <param name="d">The drive direction the lock is applied for.</param>
	void NodeHandler::addNodeLock(ui32 id, ui64 guid, driveDirection d)
	{
#ifdef TF_DEBUG_DETAIL_FULL
		std::pair<lockMap::iterator,lockMap::iterator> p = _locks.equal_range(id);
		for(lockMap::iterator ipNL = _locks.begin(); ipNL != _locks.end(); ipNL++)
		{
			if(ipNL->first == id && ipNL->second->guid == guid)
			{
				boost::format fmt = boost::format("Car: %1% already has lock on node: %2%") % guid % id;
				LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
				break;
			}
		}
#endif
		//add node Lock to list
		nodeLock* nL = new nodeLock();
		nL->direction = d;
		nL->guid = guid;
#ifdef _OPENMP
		omp_set_lock(_lock_mutex[id]);
#endif
		if(_locks[id] == NULL)
		{
			_locks[id] = new std::list<nodeLock*>();
		}
		_locks[id]->push_back(nL);
#ifdef _OPENMP
		omp_unset_lock(_lock_mutex[id]);
#endif
	}

	///<summar>
	/// Removes a lock from the current node
	///</summary>
	void NodeHandler::removeNodeLock(ui32 id, ui64 guid)
	{
#ifdef _OPENMP
		omp_set_lock(_lock_mutex[id]);
#endif
		if(_locks[id] != NULL)
		{
			for(std::list<nodeLock*>::iterator ipNL = _locks[id]->begin(); ipNL != _locks[id]->end();)
			{
				if((*ipNL)->guid == guid)
				{

					delete(*ipNL);
					ipNL = _locks[id]->erase(ipNL);
				}
				else
					ipNL++;
			}
		}
#ifdef _OPENMP
		omp_unset_lock(_lock_mutex[id]);
#endif
	}

	///<summar>
	/// Returns true if the current node is currently locked by an object and is not available for spawns or other actions
	///</summar>
	ui16 NodeHandler::nodeLockCount(ui32 id, driveDirection d)
	{
		ui16 lockCount = 0;
		if(_locks[id] != NULL)
		{
			for(std::list<nodeLock*>::iterator iStart = _locks[id]->begin(); iStart != _locks[id]->end(); iStart++)
			{
				if((*iStart)->direction == d)
					lockCount++;
			}
		}
		return lockCount;
	}

	bool NodeHandler::isNodeLocked(ui32 id)
	{
		bool result = false;
		if(_locks[id] != NULL)
			result = _locks[id]->size() > 0;
		return result;
	}
}