/*
Serial traffic simulation program for PaPo.

############################################

This application uses data from Open Street Map, henceforth
that data is made available under the Open Database License: http://opendatacommons.org/licenses/odbl/1.0/.
Any rights in individual contents ofthe database are licensed under 
the Database Contents License: http://opendatacommons.org/licenses/dbcl/1.0/
See more at: http://opendatacommons.org/licenses/odbl/#sthash.y0L9Yy1N.dpuf

############################################

Author: Martin Poppinga, Miriam von Heeremann, Marcel Ellermann
Date: 24.07.2013      
Distributed under the Boost Software License, Version 1.0.
See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)

#############################################
*/

#include "NodeHandler.h"
#include "SpawnHandler.h"
#include "boost/format.hpp"
#include "Log.h"
#include "MapHandler.h"
#include "ConfigHandler.h"
#include <mpi.h>
#include <sstream>
#include <numeric>
#include <algorithm>


namespace traffic_simulation
{
	bool SpawnHandler::_bInstance = false;
	bool SpawnHandler::_bLoaded = false;
	SpawnHandler* SpawnHandler::_instance = NULL;

	static const double _carPopMatrix[4][5] =
	{{5,-0.037554,0,2.5388491,2},
	{10,0.0837698,- 1.8198568, 11.6381330184260,-13.1654731869884521},
	{13,-0.124258219464, 4.420984438874, -50.770279194702, 194.862567523438},
	{24,0.023662683786, -1.347930787882, 24.225618753128, -130.119656917161}};

	static const double _carDestMatrix[4][5] =
	{{5, -0.01622533005100, 0, 1.2056332512, 5},
	{10, 0.01712665025504, -0.50027970, 3.7070317742294, 0.83100246174304},
	{13, 0.0133342035435, -947526/2451515, 2.56929776077242082, 4.623449173266328},
	{24, -0.003939270948, 0.287159201925, -6.1883538064, 42.57327263122}};


	/// <summary>
	/// Returns the instance of our <see cref="SpawnHandler" /> class to secure that only a single instance is created.
	/// </summary>
	/// <returns></returns>
	SpawnHandler* SpawnHandler::getInstance()
	{
		if(!_bInstance)
		{
			_instance = new SpawnHandler();
			_bInstance = true;
		}
		return _instance;
	}

	
	static void MPI_UIMAX(void *invec, void *inoutvec, int *len, MPI_Datatype *datatype)
	{
		if((*datatype) == MPI_UNSIGNED_LONG_LONG)
		{
			for(int i = 0; i < *len; i++)
			{
				((ui64*)inoutvec)[i] = ((ui64*)invec)[i] > ((ui64*)inoutvec)[i] ? ((ui64*)invec)[i] : ((ui64*)inoutvec)[i];
			}
		}
		else
		{
			MPI_Abort(MPI_COMM_WORLD, MPI_ERR_OP);
		}
	}

	/// <summary>
	/// Prevents a default instance of the <see cref="NodeHandler"/> class from being created.
	/// </summary>
	SpawnHandler::SpawnHandler()
	{
		_guidCounter = 1;
		loadSpawns();
		MPI_Op_create((MPI_User_function*)MPI_UIMAX, true, &_uimaxOp);
	}

	/// <summary>
	/// Finalizes an instance of the <see cref="NodeHandler"/> class.
	/// </summary>
	SpawnHandler::~SpawnHandler()
	{
		_bInstance = false;
		_instance = NULL;
		_bLoaded = false;
	}

	bool spawn_sort(const spawnContainer* _LEFT, const spawnContainer* _RIGHT)
	{
		return _LEFT->size > _RIGHT->size;
	}

	ui64 spawn_add(ui64 _LEFT, const spawnContainer* _RIGHT)
	{
		return _LEFT + _RIGHT->size;
	}

	void SpawnHandler::loadSpawns()
	{
		bool success = true;

		//initialize map
		for(int i = 0; i < 4; i++)
			_spawner[(SpawnerType)i] = std::vector<spawnContainer*>();

		std::stringstream inputFile;
		inputFile << world::getInputFile(FILE_SPAWNS);
		LOG->writeWorldLog("Start loading Spawns...");
		try
		{
			//open file
			std::ifstream in(inputFile.str());
			//check if file was opened
			if(in)
			{
				//go through every line
				//structure: (uInt)id,(b)trafficLight,(uInt)id
				std::string Version;
				std::getline(in,Version);
				if(CHECK_VERSION(Version.c_str()))
				{
					while(!in.eof() && in.peek() > 30 /*only writable character*/)
					{
						std::string sID, sType, sSize, sSpawnPoints;
						std::getline(in, sID, ';'); //read line till delimiter
						std::getline(in, sType, ';');
						std::getline(in, sSize, ';');
						std::getline(in, sSpawnPoints);
						spawnContainer* sc = new spawnContainer();
						sc->id = std::stoull(sID);
						sc->size = (ui32)std::stof(sSize); //floating points don't matter

						//init map to convert type to enum value
						//extract type
						std::map<std::string, SpawnerType> sTypeMap = boost::assign::map_list_of("residential",SpawnerType::residential)("industrial",SpawnerType::industrial)
							("commercial",SpawnerType::commercial)("retail",SpawnerType::retail);
						SpawnerType t =  sTypeMap.find(sType) != sTypeMap.end() ? sTypeMap[sType] : SpawnerType::residential; //default value = residential
						sc->type = t;
						//extract street / node pairs
						std::string sNSPair;
						std::stringstream pairs(sSpawnPoints);
						while(std::getline(pairs, sNSPair,':'))
						{
							ui32 NodeID;
							ui32 StreetID;
							std::stringstream v (sNSPair);
							v >> NodeID;
							v.ignore(1,',');
							v >> StreetID;
							//check distances.. should be done in preprocessing
							bool add = true;
							for(std::list<std::pair<Node*,Street*>>::iterator p = sc->spawnPoints.begin(); p != sc->spawnPoints.end(); p++)
							{
								Node* n = NODES->getNode(NodeID);
								Street* s = MAP->getStreet(StreetID);
								if(p->second->id() == StreetID)
								{
									if(fabs(s->getNodePosition(n,driveDirection::forward) - s->getNodePosition(p->first,driveDirection::forward)) < CAR_LENGTH * 2)
									{
#ifdef TF_DEBUG_DETAIL
										boost::format fmt = boost::format("Spawning node: %1% on Street: %2% has not enough distance to : %3%") % n->getID() % StreetID % p->first->getID();
										LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
#endif
										add = false;
										break;
									}
								}
							}
							if(add)
							{
								sc->spawnPoints.push_back(std::make_pair(NODES->getNode(NodeID),MAP->getStreet(StreetID)));
							}
						}
#ifdef TF_DEBUG_DETAIL
						if(sc->spawnPoints.size() == 0)
						{
							boost::format fmt = boost::format("Spawner: %1% with no spawnpoints.") % sc->id;
							LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
						}
#endif
						_spawner[t].push_back(sc);
					}
				}
				else
				{
					boost::format fmt = boost::format("Invalid File Version: %1%. Only Version: %2% supported.") % Version % MAP_DATA_VERSION;
					LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
				}
			}
		}
		catch(std::exception e)
		{
#ifdef TF_DEBUG_MAIN
			LOG->writeWorldLog("Error while loading spawns.");
			LOG->writeWorldLog(e.what());
#endif
			success = false;
		}
		_bLoaded = success;
		LOG->writeWorldLog("Finished loading spawns.");
		//sort spawns by size
		int i = 0;
#pragma omp parallel for private(i)
		for(i = 0; i < 4; i++)
		{
			std::sort(_spawner[(SpawnerType)i].begin(),_spawner[(SpawnerType)i].end(),spawn_sort);
			_sMaxSize[i] = std::accumulate(_spawner[(SpawnerType)i].begin(), _spawner[(SpawnerType)i].end(), 0LL, spawn_add);
		}
	}


	///<summary>
	/// Computes the current maximum number that should drive on the map, based on time t.
	///<param name="t">The time of the current world.</param>
	///<returns> The max number of cars for the current time</returns>
	///</summary>
	ui32 _carPopFunction(ui32 t)
	{
		double _baseT = (t % 86400) / 3600.0; //transform t into intervall [0,24)
		//get value row for our value matrix
		char mRow = 0;
		while(!(_carPopMatrix[mRow][0] > _baseT))
			mRow++;
		//compute value based on coefficients of our matrix
		double result = 0;
		for(int i = 3; i >=0; i--)
			result += _carPopMatrix[mRow][4-i] * pow(_baseT, i);
		return static_cast<int>(result * MAX_CARS / 10); //result : [0,10]
	}
	///<summary>
	/// Computes the current ratio on how many cars spawn in residential spawning areas, based on time t.
	///<param name="t">The time of the current world.</param>
	///<returns> The ratio for the current time</returns>
	///</summary>
	double _carDestFuntion(ui32 t)
	{
		ui32 _baseT = (t % 86400) / 3600; //transform t into intervall [0,24)
		//get value row for our value matrix
		char mRow = 0;
		while(!(_carDestMatrix[mRow][0] > _baseT))
			mRow++;
		//compute value based on coefficients of our matrix
		double result = 0;
		for(int i = 3; i >=0; i--)
			result += _carDestMatrix[mRow][4-i] * pow((double)_baseT, i);
		return result / 10; // result : [0,10]
	}

	/// <summary>
	/// Gets the free spawns.
	/// </summary>
	/// <param name="types">The types free spawn points are searched for.</param>
	/// <returns>A list of free spawn points and its spawnContainer</returns>
	SpawnHandler::spawnHolder SpawnHandler::getFreeSpawns(const SpawnerType* types, char arraySize)
	{
		spawnHolder returnList;
		//#pragma omp parallel for shared(returnList)
		for(int i = 0; i < arraySize; i++)
		{
			std::vector<spawnContainer*> l = SPAWNS->_spawner[types[i]];
			int j;
			ui32 max =  l.size();
#pragma omp parallel for private(j) shared(returnList, l) schedule(guided)
			for(j = 0; (ui32)j < max; j++) //go through all spawners of our current type
			{
				//check every spawn point of our spawner
				for(std::list<std::pair<Node*,Street*>>::iterator iP = l[j]->spawnPoints.begin(); iP != l[j]->spawnPoints.end(); iP++)
				{
					//check if we are responsible for the current street
					if(iP->second->getBucket() == world::getCommRank() && !NODES->isNodeLocked(iP->first->getID()))
					{
#pragma omp critical (accessReturnList)
						{
							returnList.push_back(std::make_tuple(l[j],iP->first,iP->second)); //add it to our return list
						}
					}
				}
			}
		}
		return returnList;
	}


	/// <summary>
	/// Randomly returns true, based on the propability given by cValue/mValue
	/// </summary>
	/// <param name="cValue">The c value.</param>
	/// <param name="mValue">The absolute value.</param>
	/// <returns></returns>
	bool evalRandom(ui64 cValue, ui64 mValue)
	{
		ui64 rndVal = rand();
		rndVal = static_cast<ui64>(rndVal * static_cast<long double>(mValue) / RAND_MAX);
		return rndVal < cValue;
	}

	/// <summary>
	/// Gets the spawn points. Based on propabilistic functions for car density and number of free spawn points and cars already on the streets.
	/// </summary>
	/// <param name="t_Time">The time in seconds, based at time 0 which is 03:00 am.</param>
	/// <returns>A list of nodes cars have to be spawned at</returns>
	SpawnHandler::spawnHolder SpawnHandler::getSpawnPoints(ui32 t_Time)
	{
		spawnHolder returnList;
		//max number of Cars based on current time
		int t_maxCars = _carPopFunction(t_Time);
		//get current number of Cars
		ui32 c_Cars = MAP->getCarHandler()->carCount();
		t_maxCars = t_maxCars - c_Cars; //max number of cars that can be spawned
		t_maxCars = std::min(t_maxCars, (int)SPAWN_LIMIT_PER_TICK);
		if(t_maxCars > 0 && t_maxCars < world::getCommSize() && t_maxCars >  world::getCommRank())
			t_maxCars = 1;
		else
			t_maxCars = t_maxCars / world::getCommSize();

		if(t_maxCars > 0)
		{
			double r_ResSpawn = _carDestFuntion(t_Time);
			SpawnerType rs[1] = {SpawnerType::residential};
			SpawnerType is[3] = {SpawnerType::commercial,SpawnerType::retail,SpawnerType::industrial};
			spawnHolder freeResSpawns = getFreeSpawns(rs,1);
			spawnHolder freeIndSpawns = getFreeSpawns(is,3);

			ui32 c_FreeSpawns = freeResSpawns.size() + freeIndSpawns.size(); //number of free spawning points
			ui32 c_spawn = static_cast<ui32>(t_maxCars) > c_FreeSpawns ? c_FreeSpawns : t_maxCars; //get minimum and try to spawn c_spawn cars
			ui32 c_resSpawn = static_cast<ui32>(c_spawn * r_ResSpawn);
			ui32 c_indSpawn = static_cast<ui32>(c_spawn * (1-r_ResSpawn));
			ui64 resMaxSize = 0;
			ui64 indMaxSize = 0;
			//compute accumulated size of resSpawns
			for(spawnHolder::iterator sItr = freeResSpawns.begin(); sItr != freeResSpawns.end(); sItr++)
			{
				resMaxSize += std::get<0>(*sItr)->size;
			}
			//compute accumulated size of indSpawns
			for(spawnHolder::iterator sItr = freeIndSpawns.begin(); sItr != freeIndSpawns.end(); sItr++)
			{
				indMaxSize += std::get<0>(*sItr)->size;
			}
			spawnHolder::iterator spawnItr = freeResSpawns.begin();
			ui32 added = 0;
			while(added < freeResSpawns.size() && added < c_resSpawn)
			{
				if(evalRandom(std::get<0>(*spawnItr)->size,resMaxSize))
				{
					returnList.push_back(*spawnItr);
					resMaxSize -= std::get<0>(*spawnItr)->size;
					spawnItr = freeResSpawns.erase(spawnItr);
					added++;
				}
				else
				{
					spawnItr++;
				}
				if(spawnItr == freeResSpawns.end())
				{
					spawnItr = freeResSpawns.begin();
				}
			}
			spawnItr = freeIndSpawns.begin();
			added = 0;
			while(added < freeIndSpawns.size() && added < c_indSpawn)
			{
				if(evalRandom(std::get<0>(*spawnItr)->size,resMaxSize))
				{
					returnList.push_back(*spawnItr);
					resMaxSize -= std::get<0>(*spawnItr)->size;
					spawnItr = freeIndSpawns.erase(spawnItr);
					added++;
				}
				else
				{
					spawnItr++;
				}
				if(spawnItr == freeIndSpawns.end())
				{
					spawnItr = freeIndSpawns.begin();
				}
			}
		}

		return returnList;
	}

	/// <summary>
	/// Spawns car into the world. Car is created with a start node and a path to it's goal point.
	/// When car reaches its goal point, it's removed from the world.
	/// </summary>
	void SpawnHandler::spawnCars()
	{
		spawnHolder points = getSpawnPoints(WORLD_TIME_STAMP % 86400);
#ifdef TF_DEBUG_MAIN
		if(points.size() > 5)
		{
			boost::format fmt = boost::format("Spawning %1% cars at timestep %2%") % points.size() % WORLD_TIME_STAMP;
			LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
		}
#endif
		int i;
		ui32 max = points.size();
		//synchronize guid counter
		MPI_Allreduce(MPI_IN_PLACE, &_guidCounter, 1, MPI_UNSIGNED_LONG_LONG, _uimaxOp, MPI_COMM_WORLD);
		_guidCounter += CONFIG->getUInt16("SPAWN_LIMIT_PER_TICK") * world::getCommRank();
#pragma omp parallel for private(i) shared(points) schedule(dynamic)
		for(i = 0; (ui32)i < max; i++)
		{
			std::pair<Node*,Street*> dPair;
			std::pair<Node*,Street*> sPair(std::get<1>(points[i]),std::get<2>(points[i]));
			getDestinationPoint(std::get<0>(points[i])->type, dPair);
			std::list<std::pair<Node*, Street*>*>*  path = MAP->getPath<__speedCost, __euclHeuristic>(sPair,dPair);
			if(path->size() > 0)
			{
#pragma omp critical(addCarToWorld)
				{
					MAP->getCarHandler()->addCarToWorld(_guidCounter,sPair.first, sPair.second, path);
					_guidCounter++;
				}
			}
			else
			{
				delete(path);
			}
		}
	}

	void SpawnHandler::getDestinationPoint(SpawnerType srcType, std::pair<Node*, Street*> &dPair)
	{
		SpawnerType is[3] = {SpawnerType::commercial,SpawnerType::retail,SpawnerType::industrial};
		SpawnerType dType;
		if(srcType == SpawnerType ::residential)
			dType = is[rand() % 3];
		else
			dType = SpawnerType::residential;
		ui64 iSelect = rand() * _sMaxSize[dType] / RAND_MAX;
		std::vector<spawnContainer*>::iterator itrCList = _spawner[dType].begin();
		ui64 iSumSelect = (*itrCList)->size;
		while(iSumSelect < iSelect && itrCList != _spawner[dType].end())
		{
			itrCList++;
			iSumSelect += (*itrCList)->size;
		}
		assert(itrCList != _spawner[dType].end());
		std::list<std::pair<Node*, Street*>>::iterator sItr = (*itrCList)->spawnPoints.begin();
		for(ui32 i = 0; i < rand() % (*itrCList)->spawnPoints.size(); i++)
			sItr++;
		dPair = *sItr;
	}


	void SpawnHandler::notifyPathEndReached(Car* c)
	{
#ifdef TF_DEBUG_DETAIL
		boost::format fmt = boost::format("Car: %1% reached end of path. Removing car from world.") % c->getGUID();
		LOG->writeCarLog(_FMT_TO_C_STR_(fmt));
#endif
		_delList.push_back(c);
	} 

	void SpawnHandler::update()
	{
		for(std::list<Car*>::iterator car = _delList.begin(); car != _delList.end(); car++)
		{
			MapHandler::getInstance()->getCarHandler()->removeCarFromWorld(*car);
		}
		_delList.clear();
		spawnCars();
	}

	void SpawnHandler::registerForRemove(Car* c)
	{
		_delList.push_back(c);
	}
}
