/*
Serial traffic simulation program for PaPo.

############################################

This application uses data from Open Street Map, henceforth
that data is made available under the Open Database License: http://opendatacommons.org/licenses/odbl/1.0/.
Any rights in individual contents ofthe database are licensed under 
the Database Contents License: http://opendatacommons.org/licenses/dbcl/1.0/
See more at: http://opendatacommons.org/licenses/odbl/#sthash.y0L9Yy1N.dpuf

############################################

Author: Martin Poppinga, Miriam von Heeremann, Marcel Ellermann
Date: 24.07.2013      
Distributed under the Boost Software License, Version 1.0.
See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)

#############################################
*/

#include "Street.h"
#include "NodeHandler.h"
#include "boost/format.hpp"
#include "ConfigHandler.h"
#include "Messages.h"
#include "MessageHandler.h"
#include "Log.h"
#include <sstream>

namespace traffic_simulation
{
	Street::Street() : WorldObj(0.0,0.0)
	{
		_lanes = 0;
		_length = 0;
		_delCar = false;
		_tlLock = 0;
#ifdef _OPENMP
		_street_mutex[0] = new omp_lock_t;
		omp_init_lock(_street_mutex[0]);
		_street_mutex[1] = new omp_lock_t;
		omp_init_lock(_street_mutex[1]);
#endif
	}

	Street::~Street()
	{
		//free intersections
		for(std::multimap<ui32, std::pair<Street*,std::list<Street*>>>::iterator itr = _intersections.begin(); itr != _intersections.end(); itr++)
		{
			delete(&itr->second.second); //memory allocated for Street list in [loadIntersections()]
		}
		_intersections.clear();
	}

	std::ifstream& operator>>(std::ifstream& in, Street& d)
	{
		in >> d._id;
		in.ignore(1,';');
		std::string name, type, oneway, speed, lanes, length, nodes;
		std::getline(in, name, ';');
		std::getline(in, type, ';');
		std::getline(in, oneway, ';');
		std::getline(in, speed, ';');
		std::getline(in, lanes, ';');
		std::getline(in, length, ';');
		std::getline(in, nodes);
		//init map to convert type to enum value
		//extract type
		std::map<std::string, streetType> sTypeMap = boost::assign::map_list_of("motorway",motorway)
			("motorway_link",motorway_link)("pirmary", primary)
			("primary_link", primary_link)("secondary", secondary)
			("tertiary", tertiary)("unclassified", unclassified)
			("residential",residential)("living_street",living_street)("unkown",unknown);
		std::map<std::string, streetType>::const_iterator itr = sTypeMap.find(type);
		d._name = name.c_str();
		if(itr != sTypeMap.end())
		{
			d._type = itr->second;
		}
		else
		{
			d._type = unknown;
		}
		//extract one way property
		if(oneway == "No")
			d._oneWayDirection = driveDirection::both;
		else
			d._oneWayDirection = oneway == "Forward" ? driveDirection::forward : driveDirection :: backward;

		//extract speed
		try
		{
			d._speed = speed.compare("") ? (std::stod(speed) / KM_H_TO_L_T) : 0;
			if(d._speed == 0)
				d._speed = 1;
		}
		catch(std::exception e)
		{
			bool x = true;
		}
		//check speed
#ifdef TF_DEBUG_MAIN
		if(d._speed == 0)
		{
			boost::format frmt = boost::format("Street (%1%) with invlalid Speed: %2%") % d._id % d._speed;
			LOG->writeTrafficLog(str(frmt).c_str());
		}
#endif
		//extract lines
		ui8 ilanes = lanes.compare("") ? (boost::uint8_t)std::stoi(lanes) : 0;
		d.lanes(ilanes);
		//extract length
		d._length = std::stod(length);
		std::stringstream ss(nodes);
		std::string node;

		while(std::getline(ss, node, ':'))
		{
			try
			{
				boost::uint32_t id = std::stoul(node);
				Node* n = NodeHandler::getInstance()->getNode(id);
				if(n != NULL)
				{
					d._nodes.push_back(n);
				}
				else
				{
#ifdef TF_DEBUG_MAIN
					std::stringstream sout;
					sout << "Invalid Node: " << id << " required for street: " << d._id;
					LOG->writeWorldLog(sout.str().c_str());
#endif
				}
			}
			catch(std::exception e)
			{
#ifdef TF_DEBUG_MAIN
				LOG->writeWorldLog("Error while converting node.");
				LOG->writeWorldLog(node.c_str());
#endif
			}
		}
		return in;
	}


	/// <summary>
	/// Sets the traffic light lock.
	/// </summary>
	/// <param name="d">The direction the lock is set for.</param>
	void Street::setTrafficLightLock(driveDirection d)
	{
		char mask = 0;
		if(d == driveDirection::forward)
			mask = _BYTE_MASK_FORWARD_LOCK_;
		else
			mask = _BYTE_MASK_BACKWARD_LOCK_;
		_tlLock = _tlLock | mask;
	}

	/// <summary>
	/// Removes the traffic light lock.
	/// </summary>
	/// <param name="d">The direction the lock is removed for.</param>
	void Street::removeTrafficLightLock(driveDirection d)
	{
		char mask = 0;
		if(d == driveDirection::backward)
			mask = _BYTE_MASK_FORWARD_LOCK_;
		else
			mask = _BYTE_MASK_BACKWARD_LOCK_;
		_tlLock = _tlLock & mask;
	}

	/// <summary>
	/// Determines whether a lock is set for the given direction or not.
	/// </summary>
	/// <param name="d">The drive direction.</param>
	/// <returns>True if a lock exists for the given direction</returns>
	bool Street::hasTrafficLightLock(driveDirection d)
	{
		char mask = 0;
		if(d == driveDirection::forward)
			mask = _BYTE_MASK_FORWARD_LOCK_;
		else
			mask = _BYTE_MASK_BACKWARD_LOCK_;
		return ((_tlLock & mask) > 0);
	}

	/// <summary>
	/// Determines whether a car can change on this street at the desired node or not.
	/// </summary>
	/// <param name="n">The node the car changes on this street</param>
	/// <returns>True if there is at least one free lane the car can be added to.</returns>
	bool Street::canChangeOnStreet(ui32 nodeID)
	{
#ifdef _OPENMP
		omp_set_lock(_street_mutex[0]);
		omp_set_lock(_street_mutex[1]);
#endif
		int counter = 0;
		for(std::list<std::pair<ui32, ui64>>::iterator i = _carsAtBegin.begin(); i != _carsAtBegin.end(); i++)
		{
			if(i->first == nodeID)
				counter++;
		}
#ifdef _OPENMP
		omp_unset_lock(_street_mutex[0]);
		omp_unset_lock(_street_mutex[1]);
#endif
		return counter < _lanes;
	}

	/// <summary>
	/// Determines whether the intersection between the current street and s has no prioritys to care for or not.
	/// </summary>
	/// <param name="s">The street.</param>
	/// <returns>True if we don't have any steets we need to give priority to</returns>
	bool Street::isPriorityStreet(Street* s)
	{
		bool isPstreet = true;
		for(IntersectionMap::iterator iNode = _intersections.begin(); iNode != _intersections.end(); iNode++)
		{
			if(iNode->second.first == s)
			{
				if(iNode->second.second.size() > 0)
				{
					isPstreet = false;
					break;
				}
			}
		}
		return isPstreet;
	}

	void Street::update(ui8 step)
	{
		for(CarMap::iterator lane = _cars.begin(); lane != _cars.end(); lane++)
		{
			for(std::list<Car*>::iterator car = lane->second.begin(); car != lane->second.end();)
			{
				if((*car)->getCurrentTimeStamp() == WORLD_TIME_STAMP) //only update car if it's valid for our current update cycle
					(*car)->update();
				if(_delCar)
				{
					car = removeCar(*car);	
					_delCar = false;
				}
				else
					car++;
			}
		}
	}

	void Street::addIntersection(ui32 node, Street* street, std::list<Street*> priorityWays)
	{
		/*boost::format fmt = boost::format("Adding Intersection at node: %1% for Street : %2%") % node % street->id();
		LOG->writeStreetLog(_FMT_TO_C_STR_(fmt));*/
		_intersections.insert(std::pair<ui32, std::pair<Street*,std::list<Street*>>>(node, std::pair<Street*,std::list<Street*>>(street, priorityWays)));
	}


	///<summary>
	/// Sets the lane number of our current street, increasing is always possible but be aware that lane count can only be decreased for emtpy lanes
	///</summary>
	void Street::lanes(ui8 laneCount)
	{
		if(laneCount != _lanes)
		{
			//if lane count differs adjust it
			while(_lanes < laneCount)
			{
				_cars.insert(std::make_pair(driveDirection::forward,std::list<Car*>()));
				_cars.insert(std::make_pair(driveDirection::backward,std::list<Car*>()));
				_lanes++;
			}
			std::pair<CarMap::iterator, CarMap::iterator> i = _cars.equal_range(driveDirection::forward);
			std::pair<CarMap::iterator, CarMap::iterator> j = _cars.equal_range(driveDirection::backward);
			CarMap::iterator itr2 = j.first;
			for(CarMap::iterator itr = i.first; itr != i.second;)
			{
				if(_lanes > laneCount)
				{
					if(itr->second.empty() && itr2->second.empty())
					{
						itr = _cars.erase(itr);
						itr2 = _cars.erase(itr2);
					}
					else
					{
						itr++;
						itr2++;
					}
				}
				else
				{
					itr++;
					itr2++;
				}
			}
#ifdef TF_DEBUG_DETAIL_FULL
			if(_lanes > laneCount)
			{
				boost::format fmt = boost::format("Attempt to resize lane count of street: %1%. But not enough lanes are empty.") % _id;
				LOG->writeCarLog(_FMT_TO_C_STR_(fmt));
			}
#endif
		}
	}

	/// <summary>
	/// Gets the metric node position within the street depending on the current driveDirection.
	/// </summary>
	/// <param name="n">Pointer to the Node</param>
	/// <param name="d">The drive Direction.</param>
	/// <returns>Distance from the beginning of the street to the current node (relativ to drive direction)</returns>
	double Street::getNodePosition(const Node* n, driveDirection d)
	{
		double dGN = 0;
		if(d == driveDirection::forward)
		{
			std::list<Node*>::iterator itr = _nodes.begin();
			std::list<Node*>::iterator itr2 = std::list<Node*>::iterator(itr);
			itr2++;
			while(*itr != n && itr2 != _nodes.end())
			{
				dGN += (*itr)->getPos()->distance((*itr2)->getPos());
				itr++;
				itr2++;
			}
		}
		if(d == driveDirection::backward)
		{
			std::list<Node*>::reverse_iterator itr = _nodes.rbegin();
			std::list<Node*>::reverse_iterator itr2 = std::list<Node*>::reverse_iterator(itr);
			itr2++;
			while(*itr != n && itr2 != _nodes.rend())
			{
				dGN += (*itr)->getPos()->distance((*itr2)->getPos());
				itr++;
				itr2++;
			}
		}
		return dGN;
	}

	const Car* Street::getFrontCar(const Car* c)
	{
		CarMap::iterator lane;
		std::list<Car*>::iterator car;
		findCar(c, car, lane);
		for(std::list<Car*>::reverse_iterator c2 = lane->second.rbegin(); c2 != lane->second.rend(); c2++)
		{
			if((*c2)->getGUID() == c->getGUID())
			{
				c2++;
				if(c2 != lane->second.rend())
					return *c2;
				else
					break;
			}
		}
		return NULL;
	}
	///<summar>
	/// Spawns a car on the street - use this to add cars to the world
	/// Returns the Metric Position on this street
	///</summar>
	bool Street::spawnCar(Car* c, driveDirection d)
	{
		if(isOneWay() && _oneWayDirection != d)
		{
#ifdef TF_DEBUG_DETAIL_FULL
			boost::format fmt = boost::format("Car(%1%) attempts to drive into invalid direction on street: %2%") % c->getGUID() % this->_id;
			LOG->writeStreetLog(_FMT_TO_C_STR_(fmt));
#endif
		}
		else
		{
			//improve spawning by using all lanes
			std::pair<CarMap::iterator, CarMap::iterator> iP = _cars.equal_range(d);
			for(CarMap::iterator lane = iP.first; lane != iP.second; lane++)
			{
				if(lane->second.size() > 0) //no distance checks here, we believe that our locking system works properly and therefore only free nodes with enough distance are chosen
				{
					for(std::list<Car*>::iterator car = lane->second.begin(); car != lane->second.end(); car++)
					{
						if((*car)->getMetricPos() < c->getMetricPos())
						{
#ifdef _OPENMP
							omp_set_lock(_street_mutex[lane->first]);
#endif
							lane->second.insert(car, c);
#ifdef _OPENMP
							omp_unset_lock(_street_mutex[lane->first]);
#endif
							return true;
						}
					}
#ifdef _OPENMP
					omp_set_lock(_street_mutex[lane->first]);
#endif	
					lane->second.push_back(c);
#ifdef _OPENMP
					omp_unset_lock(_street_mutex[lane->first]);
#endif
					return true;
				} //no car on lane just spawn
				else
				{
#ifdef _OPENMP
					omp_set_lock(_street_mutex[lane->first]);
#endif
					lane->second.push_back(c);
#ifdef _OPENMP
					omp_unset_lock(_street_mutex[lane->first]);
#endif
					return true;
				}
			}
		}
#ifdef TF_DEBUG_DETAIL
		boost::format fmt = boost::format("Unable to spawn car: %1% on street %2% driveDirection: %3%") % c->getGUID() % _id %
			(d == driveDirection::forward ? "Forward" : "Backward");
		LOG->writeStreetLog(_FMT_TO_C_STR_(fmt));
#endif
		return false;
	}


	/// <summary>
	/// Returns a free lane a car can be added to.
	/// </summary>
	/// <param name="n">The node the car should be added at.</param>
	/// <returns>CarMap iterator pointing to a lane if there is a free lane, or pointing to _the map end if there is no free lane</returns>
	Street::CarMap::iterator Street::getFreeLane(ui32 nodeID)
	{
		CarMap::iterator i = _cars.end();
		if((*_nodes.begin())->getID() == nodeID)
		{
#ifdef _OPENMP
			omp_lock_t* lock = _street_mutex[driveDirection::forward];
			omp_set_lock(lock);
#endif
			std::pair<CarMap::iterator, CarMap::iterator> forwardLanes = _cars.equal_range(driveDirection::forward);
			for(CarMap::iterator lane = forwardLanes.first; lane != forwardLanes.second; lane++)
			{
				if(lane->second.empty() || (*lane->second.rbegin())->getMetricPos() > KM_H_TO_L_T * _speed / 2 + _speed)
				{
					i = lane;
					break;
				}
			}
#ifdef _OPENMP
			omp_unset_lock(_street_mutex[driveDirection::forward]);
#endif
		}
		else if((*_nodes.rbegin())->getID() == nodeID)
		{
#ifdef _OPENMP
			omp_set_lock(_street_mutex[driveDirection::backward]);
#endif
			std::pair<CarMap::iterator, CarMap::iterator> backwardLanes = _cars.equal_range(driveDirection::backward);
			for(CarMap::iterator lane = backwardLanes.first; lane != backwardLanes.second; lane++)
			{

				if(lane->second.empty() || (*lane->second.rbegin())->getMetricPos() > KM_H_TO_L_T * _speed / 2 + _speed)
				{
					i = lane;
					break;
				}
			}
#ifdef _OPENMP
			omp_unset_lock(_street_mutex[driveDirection::backward]);
#endif
		}
		return i;
	}
	///<summar>
	/// Adds a car to one end of the street. Be sure not to use this when u spawn cars, becaus spawning nodes might not be at the beginning / end of a street.
	///</summar>
	void Street::addCar(Car* c, ui32 nodeID)
	{
		CarMap::iterator i = getFreeLane(nodeID);
		if(i != _cars.end())
		{
#pragma omp critical (addCarToLane)
			{
				i->second.push_back(c);
			}
		}
#ifdef TF_DEBUG_MAIN
		else
		{
			boost::format fmt = boost::format("Couldn't add car(%1%) to street : %2%") % c->getGUID() % _id;
			LOG->writeStreetLog(_FMT_TO_C_STR_(fmt));
		}
#endif
	}

	std::list<Car*>::iterator Street::removeCar(const Car* c)
	{
#ifdef TF_DEBUG_DETAIL_FULL
		boost::format fmt = boost::format("Removing car: %1% on street %2%") % c->getGUID() % _id;
		LOG->writeStreetLog(_FMT_TO_C_STR_(fmt));
#endif
		bool carRemoved = false;
		CarMap::iterator lane;
		std::list<Car*>::iterator car;
		findCar(c, car, lane);
		if(lane != _cars.end())
		{
			notifyNotCloseToEnd((*car)->getGUID());
			notifyNotCloseToBegin((*car)->getGUID());
#ifdef _OPENMP
			omp_set_lock(_street_mutex[lane->first]);
#endif
			car = lane->second.erase(car);
#ifdef _OPENMP
			omp_unset_lock(_street_mutex[lane->first]);
#endif
			carRemoved = true;

		}
#ifdef TF_DEBUG_MAIN
		if(!carRemoved)
		{

			boost::format fmt = boost::format("Invalid GUID requested to be deleted. GUID: %1% Street: %2%") % c->getGUID() % this->id();
			LOG->writeStreetLog(_FMT_TO_C_STR_(fmt));
		}
#endif
		return car;
	}


	template<typename T>
	bool listContainsValue(std::list<T>* pL, T v)
	{
		bool result = false;
		for(typename std::list<T>::const_iterator i = pL->cbegin(); i != pL->cend(); i++)
		{
			if(*i == v)
			{
				result = true;
				break;
			}
		}
		return result;
	}

	void Street::notifyNotCloseToBegin(ui64 carID)
	{
#ifdef _OPENMP
		omp_set_lock(_street_mutex[0]);
		omp_set_lock(_street_mutex[1]);
#endif
		for(std::list<std::pair<ui32,ui64>>::iterator i = _carsAtBegin.begin(); i != _carsAtBegin.end();)
		{
			if(i->second == carID)
			{
				std::list<Street*> _sAtInter = getIntersectingStreetsAtNode(i->first);
				for(std::list<Street*>::iterator sItr = _sAtInter.begin(); sItr != _sAtInter.end(); sItr++)
				{
					if((*sItr)->getBucket() != world::getCommRank())
					{	
						__NotCloseToBeginStoreStruct nctobStruct;
						nctobStruct._guid = carID;
						nctobStruct._streetID = _id;
						MSG->SendMessage(&nctobStruct, 1, MSG_NOT_CLOSE_TO_BEGIN, (*sItr)->getBucket());
					}
				}
				i = _carsAtBegin.erase(i);
				break;
			}
			else
				i++;
		}
#ifdef _OPENMP
		omp_unset_lock(_street_mutex[0]);
		omp_unset_lock(_street_mutex[1]);
#endif
	}
	void Street::notifyCloseToBegin(ui64 carID, ui32 nodeID)
	{
		driveDirection d = driveDirection::forward;
		if(nodeID != (*_nodes.begin())->getID())
			d = driveDirection::backward;
#ifdef _OPENMP
		omp_set_lock(_street_mutex[d]);
#endif
		if(!listContainsValue<std::pair<ui32,ui64>>(&_carsAtBegin, std::make_pair(nodeID, carID)))
		{
			_carsAtBegin.push_back(std::make_pair(nodeID, carID));
			std::list<Street*> _sAtInter = getIntersectingStreetsAtNode(nodeID);
			for(std::list<Street*>::iterator sItr = _sAtInter.begin(); sItr != _sAtInter.end(); sItr++)
			{
				if((*sItr)->getBucket() != world::getCommRank())
				{
					__CloseToBeginStoreStruct ctbStore;
					ctbStore._guid = carID;
					ctbStore._nodeID = nodeID;
					ctbStore._streetID = _id;
					MSG->SendMessage(&ctbStore, 1, MSG_CLOSE_TO_BEGIN, (*sItr)->getBucket());
				}
			}
		}
#ifdef _OPENMP
		omp_unset_lock(_street_mutex[d]);
#endif		
	}

	std::list<Street*> Street::getIntersectingStreetsAtNode(ui32 nodeID)
	{
		std::list<Street*> rList;
		std::pair<IntersectionMap::iterator, IntersectionMap::iterator> iPair = _intersections.equal_range(nodeID);
		for(IntersectionMap::iterator i = iPair.first; i != iPair.second; i++)
		{
			rList.push_back(i->second.first);
		}
		return rList;
	}


	void Street::notifyNotCloseToEnd(const ui64 carID)
	{
#ifdef _OPENMP
		omp_set_lock(_street_mutex[0]);
		omp_set_lock(_street_mutex[1]);
#endif
		for(std::list<std::tuple<ui64,ui32,ui32>>::iterator itr = _priority.begin(); itr != _priority.end();)
		{
			if(std::get<0>(*itr) == carID)
			{
				std::list<Street*> _sAtInter = getIntersectingStreetsAtNode(std::get<1>(*itr));
				for(std::list<Street*>::iterator sItr = _sAtInter.begin(); sItr != _sAtInter.end(); sItr++)
				{
					if((*sItr)->getBucket() != world::getCommRank())
					{
						__NotCloseToEndStoreStruct ncteStorage;
						ncteStorage._guid = carID;
						ncteStorage._CstreetID = _id;
						MSG->SendMessage(&ncteStorage, 1, MSG_NOT_CLOSE_TO_END, (*sItr)->getBucket());
					}
				}
				itr = _priority.erase(itr);
				break;
			}
			else
			{
				itr++;
			}
		}
#ifdef _OPENMP
		omp_unset_lock(_street_mutex[0]);
		omp_unset_lock(_street_mutex[1]);
#endif
	}
	/// <summary>
	/// Notifies a street that a car is close to it's end
	/// </summary>
	/// <param name="c">The car that is close to the end.</param>
	void Street::notifyCloseToEnd(ui64 carID, ui32 nodeID, ui32 streetID)
	{
#ifdef _OPENMP
		omp_set_lock(_street_mutex[0]);
		omp_set_lock(_street_mutex[1]);
#endif
		if(!listContainsValue<std::tuple<ui64, ui32, ui32>>(&_priority, std::make_tuple(carID,nodeID,streetID)))
		{
			_priority.push_back(std::make_tuple(carID,nodeID,streetID));
			std::list<Street*> _sAtInter = getIntersectingStreetsAtNode(nodeID);
			__CloseToEndStoreStruct _ctoStorage;
			_ctoStorage._guid = carID;
			_ctoStorage._CstreetID = this->_id;
			_ctoStorage._GnodeID = nodeID;
			_ctoStorage._GstreetID = streetID;
			for(std::list<Street*>::iterator iS = _sAtInter.begin(); iS != _sAtInter.end(); iS++)
			{
				if((*iS)->getBucket() != world::getCommRank()) //if this street connects to streets that belong to other processes, notify'em
					MSG->SendMessage(&_ctoStorage, 1, MSG_CLOSE_TO_END, (*iS)->getBucket());
			}
		}
#ifdef _OPENMP
		omp_unset_lock(_street_mutex[0]);
		omp_unset_lock(_street_mutex[1]);
#endif
	}

	/// <summary>
	/// Finds the car.
	/// </summary>
	/// <param name="c">Pointer to our car.</param>
	/// <param name="i">Iterator that is set to the lane list the car is in.</param>
	/// <param name="j">Iterator that is set to the key value pair of the lane in our map.</param>
	void Street::findCar(const Car* c, std::list<Car*>::iterator &i, CarMap::iterator &j)
	{
		for(CarMap::iterator itr = _cars.begin(); itr != _cars.end(); itr++)
		{
			for(std::list<Car*>::iterator car = itr->second.begin(); car != itr->second.end(); car++)
			{
				if((*car)->getGUID() == c->getGUID())
				{
					j = itr;
					i = car;
					return;
				}
			}
		}
		j = _cars.end();
	}


	/// <summary>
	/// Determines whether the street has a priority care to care for or not.
	/// </summary>
	/// <param name="nodeID">The node identifier.</param>
	/// <returns>True if there is a proiority car on this street for our direction</returns>
	bool Street::hasPriorityCar(ui32 nodeID)
	{
		bool hasPriority = false;
		//iterate over all cars that have called for priority
#ifdef _OPENMP
		omp_set_lock(_street_mutex[0]);
		omp_set_lock(_street_mutex[1]);
#endif
		for(std::list<std::tuple<ui64,ui32,ui32>>::iterator car = _priority.begin(); car != _priority.end(); car++)
		{
			if(std::get<1>(*car) == nodeID)
			{
				hasPriority = true;
				break;
			}
		}
#ifdef _OPENMP
		omp_unset_lock(_street_mutex[0]);
		omp_unset_lock(_street_mutex[1]);
#endif
		return hasPriority;
	}

	std::set<Street*> Street::getPriorityStreets(ui32 nodeID, ui32 streetID)
	{
		std::set<Street*> rList;
		std::pair<IntersectionMap::iterator,IntersectionMap::iterator> pIPrio = _intersections.equal_range(nodeID);
		for(IntersectionMap::iterator i = pIPrio.first; i != pIPrio.second; i++)
		{
			if(i->second.first->id() == streetID)
			{
				for(std::list<Street*>::iterator iStreet = i->second.second.begin(); iStreet != i->second.second.end(); iStreet++)
				{
					if((*iStreet)->hasPriorityCar(nodeID))
					{
						rList.insert(*iStreet);
					}
				}
				break;
			}
		}
		return rList;
	}

	std::list<std::tuple<ui64,ui32,ui32>> Street::getPriorityCars(ui32 nodeID)
	{
		std::list<std::tuple<ui64,ui32,ui32>> cList;
#ifdef _OPENMP
		omp_set_lock(_street_mutex[0]);
		omp_set_lock(_street_mutex[1]);
#endif
		for(std::list<std::tuple<ui64,ui32,ui32>>::const_iterator car = _priority.begin(); car != _priority.end(); car++)
		{
			if(std::get<1>(*car) == nodeID)
				cList.push_back(*car);
		}
#ifdef _OPENMP
		omp_unset_lock(_street_mutex[0]);
		omp_unset_lock(_street_mutex[1]);
#endif
		return cList;
	}
}
