/*
	Serial traffic simulation program for PaPo.

	############################################

	This application uses data from Open Street Map, henceforth
	that data is made available under the Open Database License: http://opendatacommons.org/licenses/odbl/1.0/.
	Any rights in individual contents ofthe database are licensed under 
	the Database Contents License: http://opendatacommons.org/licenses/dbcl/1.0/
	See more at: http://opendatacommons.org/licenses/odbl/#sthash.y0L9Yy1N.dpuf

	############################################

	Author: Martin Poppinga, Miriam von Heeremann, Marcel Ellermann
	Date: 24.07.2013      
	Distributed under the Boost Software License, Version 1.0.
	    See accompanying file LICENSE_1_0.txt or copy at
	          http://www.boost.org/LICENSE_1_0.txt)

	#############################################
*/

#pragma once

#include "WorldObj.h"
#include "worldDefs.h"
#include "boost/assign/list_of.hpp"
#include "NodeHandler.h"
#include "CarHandler.h"
#include "omp.h"
#include "Car.h"
#include <map>
#include <list>
#include <set>

namespace traffic_simulation
{

#define _BYTE_MASK_FORWARD_LOCK_ 1
#define _BYTE_MASK_BACKWARD_LOCK_ 2

	enum streetType /*From OpenStreet Wiki definitions*/
	{
		motorway /*Autobahn*/, motorway_link /*Auffahrten*/, primary /*Hauptstrae*/,
		primary_link, secondary, tertiary, residential, living_street,  unclassified, unknown
	};


	/// <summary>
	/// Structure from preprocessing documentation
	///	wayid;Straenname;type;oneway;speed;lines;lnge;Nodeid1:Nodeid2:...:Nodeidn
	/// </summary>
	class Street : public WorldObj
	{

	public:
		typedef std::multimap<ui32, std::pair<Street*,std::list<Street*>>> IntersectionMap;
		typedef std::multimap<driveDirection, std::list<Car*>> CarMap;
		Street();
		~Street();
		/*field accessor*/
		bool isOneWay(){return (_oneWayDirection != driveDirection::both);};
		ui32 id(){return _id;};
		ui8 lanes(){return _lanes;};
		void lanes(ui8 laneCount);
		double speed(){return _speed;};
		double length(){return _length;};
		const char* name(){return _name;};
		driveDirection oneWayDirection(){return _oneWayDirection;};
		void setBucket(int bID){_bucket = bID;};
		int  getBucket(){return _bucket;};
		/*navigation fields*/
		void notifyCloseToEnd(ui64 carID, ui32 nodeID, ui32 streetID);
		void notifyCloseToBegin(ui64 carID, ui32 nodeID);
		void notifyNotCloseToBegin(ui64 carID);
		void notifyNotCloseToEnd(const ui64 carID);
		const Car* getFrontCar(const Car* c);
		double getNodePosition(const Node* n, driveDirection d);
		std::list<std::tuple<ui64, ui32, ui32>> getPriorityCars(ui32 nodeID);
		bool hasPriorityCar(ui32 nodeID);
		bool isPriorityStreet(Street* s);
		bool canChangeOnStreet(ui32 nodeID);
		void addCar(Car* c, ui32 nodeID);
		void update(ui8 step = 1);
		void setTrafficLightLock(driveDirection d);
		bool hasTrafficLightLock(driveDirection d);
		void removeTrafficLightLock(driveDirection d);
		void setCarRemove(){_delCar = true;};
		bool spawnCar(Car* c, driveDirection d);
		void setID(ui32 id){_id = id;};
		void addIntersection(ui32 node, Street* street, std::list<Street*> priorityWays);
		const std::list<Node*>* nodes(){return &_nodes;};
		const std::multimap<ui32, std::pair<Street*,std::list<Street*>>>* intersections(){return &_intersections;};
		std::set<Street*> getPriorityStreets(ui32 nodeID, ui32 streetID);
		friend class CarHandler;
		//set sorting method as friend to set bucket and change the street id
	private:
		friend std::ifstream& operator>>(std::ifstream& in, Street& d);
		streetType _type;
		double _length;
		driveDirection _oneWayDirection;
		ui8 _lanes;
		double _speed;
		ui32 _id;
		std::list<std::tuple<ui64, ui32, ui32>> _priority;
		std::list<std::pair<ui32,ui64>> _carsAtBegin;
		char _tlLock;
		bool _delCar;
		const char* _name;
		int _bucket;
		std::list<Node*> _nodes;
		std::list<Street*> getIntersectingStreetsAtNode(ui32 node);
		omp_lock_t* _street_mutex[2];
		CarMap::iterator getFreeLane(ui32 nodeID);
		std::list<Car*>::iterator removeCar(const Car* c);
		void findCar(const Car* c, std::list<Car*>::iterator &i, CarMap::iterator &j);
		///list of lines for each direction with respectiv cars
		CarMap _cars;
		IntersectionMap _intersections;
	};
}