/*
Serial traffic simulation program for PaPo.

############################################

This application uses data from Open Street Map, henceforth
that data is made available under the Open Database License: http://opendatacommons.org/licenses/odbl/1.0/.
Any rights in individual contents ofthe database are licensed under 
the Database Contents License: http://opendatacommons.org/licenses/dbcl/1.0/
See more at: http://opendatacommons.org/licenses/odbl/#sthash.y0L9Yy1N.dpuf

############################################

Author: Martin Poppinga, Miriam von Heeremann, Marcel Ellermann
Date: 24.07.2013      
Distributed under the Boost Software License, Version 1.0.
See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)

#############################################
*/

#include "TrafficLightHandler.h"
#include <fstream>
#include "Log.h"
#include "boost/format.hpp"
#include <sstream>

#define TRAFFIC_STATE_GREEN 1
#define TRAFFIC_STATE_RED 0


namespace traffic_simulation
{
	/*
	TrafficLightSystemFactory
	*/
	TrafficLightSystem* TrafficLightSystemFactory::createTrafficLightSystem(ui8 tlCount)
	{
		TrafficLightSystem* tlSystem = NULL;
		//handle creation here
		switch(tlCount)
		{
		case 1:
			tlSystem = new SingleLightSystem();
			break;
		case 3:
			tlSystem = new ThreeLightSystem();
			break;
		case 4:
			tlSystem = new FourLightSystem();
			break;
		default :
			boost::format fmt = boost::format("Unknown Traffic Light System requested. Traffic Light count : %1%") % tlCount;
			LOG->writeTrafficLog(_FMT_TO_C_STR_(fmt));
		}

		return tlSystem;
	}
	/*
	TrafficLightHandler Code
	*/
	//init static variables
	bool TrafficLightHandler::_bInstance = false;
	bool TrafficLightHandler::_bLoaded = false;
	TrafficLightHandler* TrafficLightHandler::_instance = NULL;

	/// <summary>
	/// Returns the instance of our <see cref="TrafficLightHandler" /> class to secure that only a single instance is created.
	/// </summary>
	/// <returns></returns>
	TrafficLightHandler* TrafficLightHandler::getInstance()
	{
		if(!_bInstance)
		{
			_instance = new TrafficLightHandler();
			_bInstance = true;
		}
		return _instance;
	}

	/// <summary>
	/// Prevents a default instance of the <see cref="TrafficLightHandler"/> class from being created.
	/// </summary>
	TrafficLightHandler::TrafficLightHandler()
	{
		loadTrafficLights();
		_bLoaded = true;
	}

	/// <summary>
	/// Finalizes an instance of the <see cref="TrafficLightHandler"/> class.
	/// </summary>
	TrafficLightHandler::~TrafficLightHandler()
	{
		_bInstance = false;
		_instance = NULL;
		_bLoaded = false;

		for(std::list<TrafficLightSystem*>::iterator itr = _tlSystems.begin(); itr != _tlSystems.end();)
		{
			delete(*itr);
			itr = _tlSystems.erase(itr);
		}
	}

	void TrafficLightHandler::update(ui8 t_step)
	{
		for(std::list<TrafficLightSystem*>::iterator itr = _tlSystems.begin(); itr != _tlSystems.end(); itr++)
		{
			(*itr)->update(t_step);
		}
	}

	std::ifstream& operator >>(std::ifstream& in, TrafficLightSystem** tl)
	{
		//read dataline
		std::string line;
		std::getline(in,line);
		std::string tLight;
		std::stringstream input(line);
		std::list<TrafficLight*> tls;
		while(std::getline(input,tLight,':'))
		{
			std::stringstream ssLSystem(tLight);
			TrafficLight* tmpLight = new TrafficLight();
			tmpLight->State= false;
			std::string nodeID;
			std::getline(ssLSystem, nodeID, ',');
			tmpLight->nodeID = std::stoul(nodeID);
			std::string streetID;
			while(std::getline(ssLSystem, streetID, ','))
			{
				tmpLight->regulatedStreets.push_back(MAP->getStreet(std::stoul(streetID)));
			}
			tls.push_back(tmpLight);
		}
		*tl = TrafficLightSystemFactory::createTrafficLightSystem((ui8)tls.size());
		for(std::list<TrafficLight*>::iterator tmpLight = tls.begin(); tmpLight != tls.end(); tmpLight++)
		{
			if(*tl != NULL)
			{
				(*tl)->addTrafficLight(*tmpLight);
			}
			else
				delete *tmpLight; //free allocated memory
		}
		return in;
	}

	void TrafficLightHandler::loadTrafficLights()
	{
		std::string fileName = world::getInputFile(FILE_CROSSROADS);
		LOG->writeWorldLog("Start loading Traffic Lights...");
		std::ifstream in(fileName.c_str());
		if(in)
		{
			std::string Version;
			std::getline(in,Version);
			if(CHECK_VERSION(Version.c_str()))
			{
				try
				{
					while(!in.eof() && in.peek() > 30 /*only writable character*/)
					{
						TrafficLightSystem* tls = NULL;
						in >> &tls;
						if(tls != NULL)
						{
							tls->initSystemStates();
							_tlSystems.push_back(tls);
						}
					}
				}
				catch(std::exception e)
				{
					LOG->writeWorldLog("Error while loading Traffic Lights.");
					LOG->writeWorldLog(e.what());
				}
			}
			else
			{
				boost::format fmt = boost::format("Invalid File Version: %1%. Only Version: %2% supported.") % Version % MAP_DATA_VERSION;
				LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
			}
		}
		else
		{
			boost::format  fmt = boost::format("Unable to load %1%") % fileName;
			LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
		}
		LOG->writeWorldLog("Finished loading Traffic Lights");
	}


	/*
	TrafficLightSystem Code
	*/
	TrafficLightSystem::~TrafficLightSystem()
	{
		//free allocated Memory
		for(std::vector<TrafficLight*>::iterator iTL = _lights.begin(); iTL != _lights.end();)
		{
			(*iTL)->regulatedStreets.clear();
			delete(*iTL);
			iTL = _lights.erase(iTL);
		}
	}

	/// <summary>
	/// Updates the trafficlight system. Internal states are changed if enough time passed.
	/// </summary>
	/// <param name="t_step">How many LT steps have passed.</param>
	void TrafficLightSystem::update(ui8 t_step)
	{
		_internalState += t_step;
		updateTrafficLightStates();
	}

	/// <summary>
	/// Adds the traffic light.
	/// </summary>
	/// <param name="TL">Pointer to trafficlight.</param>
	void TrafficLightSystem::addTrafficLight(TrafficLight* TL)
	{
		_lights.push_back(TL);
	}

	void TrafficLightSystem::setStreetLocks(TrafficLight* tl)
	{
		for(std::list<Street*>::iterator iStreet = tl->regulatedStreets.begin(); iStreet != tl->regulatedStreets.end(); iStreet++)
		{
			if(tl->nodeID == (*(*iStreet)->nodes()->begin())->getID())
			{
				tl->State == TRAFFIC_STATE_GREEN ? (*iStreet)->removeTrafficLightLock(driveDirection::backward) :
					(*iStreet)->setTrafficLightLock(driveDirection::backward);
			}
			if(tl->nodeID == (*(*iStreet)->nodes()->rbegin())->getID())
			{
				tl->State == TRAFFIC_STATE_GREEN ? (*iStreet)->removeTrafficLightLock(driveDirection::forward) :
					(*iStreet)->setTrafficLightLock(driveDirection::forward);
			}
		}
	}

	/// <summary>
	/// Initializes the system states.
	/// </summary>
	void SingleLightSystem::initSystemStates()
	{
		for(std::vector<TrafficLight*>::iterator itr = _lights.begin(); itr != _lights.end(); itr++)
		{
			(*itr)->State = TRAFFIC_STATE_GREEN;
			setStreetLocks(*itr);
			phase = 0;
		}
	}

	/// <summary>
	/// Updates the traffic light states.
	/// </summary>
	void SingleLightSystem::updateTrafficLightStates()
	{
		if(phase == 0 && _internalState >= 120)
		{
#ifdef TF_DEBUG_DETAIL_FULL
			boost::format fmt = boost::format("SingleLightSytstem: Entering Red Phase. Node: %1%") % _lights[0]->nodeID;
			LOG->writeTrafficLog(_FMT_TO_C_STR_(fmt));
#endif
			_lights[0]->State = TRAFFIC_STATE_RED;
			setStreetLocks(_lights[0]);
			phase = 1;
			_internalState = 0;
		}
		else if(phase == 1 && _internalState == 20)
		{
#ifdef TF_DEBUG_DETAIL_FULL
			boost::format fmt = boost::format("SingleLightSytstem: Entering Green Phase. Node: %1%") % _lights[0]->nodeID;
			LOG->writeTrafficLog(_FMT_TO_C_STR_(fmt));
#endif
			_lights[0]->State = TRAFFIC_STATE_GREEN;
			setStreetLocks(_lights[0]);
			phase = 0;
			_internalState = 0;
		}
	}


	void ThreeLightSystem::initSystemStates()
	{
		_lights[0]->State = TRAFFIC_STATE_GREEN;
		setStreetLocks(_lights[0]);
		_lights[1]->State = TRAFFIC_STATE_GREEN;
		setStreetLocks(_lights[1]);
		_lights[2]->State = TRAFFIC_STATE_RED;
		setStreetLocks(_lights[2]);
	}

	void ThreeLightSystem::updateTrafficLightStates()
	{
		if(phase == 0 && _internalState >= 60)
		{
#ifdef TF_DEBUG_DETAIL_FULL
			boost::format fmt = boost::format("ThreeLightSytstem: Entering Red Phase. Node: %1%") % _lights[0]->nodeID;
			LOG->writeTrafficLog(_FMT_TO_C_STR_(fmt));
#endif
			_lights[0]->State = TRAFFIC_STATE_RED;
			setStreetLocks(_lights[0]);
			_lights[1]->State = TRAFFIC_STATE_RED;
			setStreetLocks(_lights[1]);
			phase = 1;
			_internalState = 0;
		}
		else if(phase == 1 && _internalState >= 7)
		{
#ifdef TF_DEBUG_DETAIL_FULL
			boost::format fmt = boost::format("ThreeLightSytstem: Entering Green Phase. For single Light Node: %1%") % _lights[0]->nodeID;
			LOG->writeTrafficLog(_FMT_TO_C_STR_(fmt));
#endif
			_lights[2]->State = TRAFFIC_STATE_GREEN;
			setStreetLocks(_lights[2]);
			phase = 2;
			_internalState = 0;
		}
		else if(phase == 2 && _internalState >= 20)
		{
			_lights[2]->State = TRAFFIC_STATE_RED;
			setStreetLocks(_lights[2]);
			phase = 3;
			_internalState = 0;
		}
		else if(phase == 3 && _internalState >= 5)
		{
			_lights[0]->State = TRAFFIC_STATE_GREEN;
			_lights[1]->State = TRAFFIC_STATE_GREEN;
			setStreetLocks(_lights[1]);
			setStreetLocks(_lights[0]);
			phase = 0;
			_internalState = 0;
		}
	}

	void FourLightSystem::initSystemStates()
	{
		_lights[0]->State = TRAFFIC_STATE_GREEN;
		setStreetLocks(_lights[0]);
		_lights[1]->State = TRAFFIC_STATE_GREEN;
		setStreetLocks(_lights[1]);
		_lights[2]->State = TRAFFIC_STATE_RED;
		setStreetLocks(_lights[2]);
		_lights[3]->State = TRAFFIC_STATE_RED;
		setStreetLocks(_lights[3]);
	}

	void FourLightSystem::updateTrafficLightStates()
	{
		if(phase == 0 && _internalState >= 30)
		{
#ifdef TF_DEBUG_DETAIL_FULL
			boost::format fmt = boost::format("ThreeLightSytstem: Entering Red Phase. Node: %1%") % _lights[0]->nodeID;
			LOG->writeTrafficLog(_FMT_TO_C_STR_(fmt));
#endif
			_lights[0]->State = TRAFFIC_STATE_RED;
			setStreetLocks(_lights[0]);
			_lights[1]->State = TRAFFIC_STATE_RED;
			setStreetLocks(_lights[1]);
			phase = 1;
			_internalState = 0;
		}
		else if(phase == 1 && _internalState >= 8)
		{
#ifdef TF_DEBUG_DETAIL_FULL
			boost::format fmt = boost::format("ThreeLightSytstem: Entering Green Phase. For single Light Node: %1%") % _lights[0]->nodeID;
			LOG->writeTrafficLog(_FMT_TO_C_STR_(fmt));
#endif
			_lights[2]->State = TRAFFIC_STATE_GREEN;
			_lights[3]->State = TRAFFIC_STATE_GREEN;
			setStreetLocks(_lights[2]);
			setStreetLocks(_lights[3]);
			phase = 2;
			_internalState = 0;
		}
		else if(phase == 2 && _internalState >= 30)
		{
			_lights[2]->State = TRAFFIC_STATE_RED;
			_lights[2]->State = TRAFFIC_STATE_RED;
			setStreetLocks(_lights[2]);
			setStreetLocks(_lights[3]);
			phase = 3;
			_internalState = 0;
		}
		else if(phase == 3 && _internalState >= 8)
		{
			_lights[0]->State = TRAFFIC_STATE_GREEN;
			_lights[1]->State = TRAFFIC_STATE_GREEN;
			setStreetLocks(_lights[1]);
			setStreetLocks(_lights[0]);
			phase = 0;
			_internalState = 0;
		}
	}
}
