/*
Serial traffic simulation program for PaPo.

############################################

This application uses data from Open Street Map, henceforth
that data is made available under the Open Database License: http://opendatacommons.org/licenses/odbl/1.0/.
Any rights in individual contents ofthe database are licensed under 
the Database Contents License: http://opendatacommons.org/licenses/dbcl/1.0/
See more at: http://opendatacommons.org/licenses/odbl/#sthash.y0L9Yy1N.dpuf

############################################

Author: Martin Poppinga, Miriam von Heeremann, Marcel Ellermann
Date: 24.07.2013      
Distributed under the Boost Software License, Version 1.0.
See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)

#############################################
*/

#include "world.h"
#include "NodeHandler.h"
#include "MessageHandler.h"
#include "CarHandler.h"
#include "MapHandler.h"
#include "SpawnHandler.h"
#include "ConfigHandler.h"
#include "TrafficLightHandler.h"
#include "Log.h"
#include "mpi.h"
#include "boost/format.hpp"
#include <string>

namespace traffic_simulation
{

	bool world::_bInstance = false;
	bool world::_bLoaded = false;
	world* world::_instance = NULL;
	boost::uint32_t world::_lTimeStamp = 0;
	std::string world::_fileName = "";
	int world::_rank = -1;
	int world::_size = -1;

	int world::getCommRank()
	{
		if(_rank == -1)
		{
			MPI_Comm_rank(MPI_COMM_WORLD, &_rank);
		}
		return _rank;
	}

	int world::getCommSize()
	{
		if(_size == -1)
		{
			MPI_Comm_size(MPI_COMM_WORLD, &_size);
		}
		return _size;
	}
	///<summar>
	///Constructor for our world.
	///World loads and initializes our world objects and loads the Object Handlers for Map and Car handling.
	///</summar>
	world::world()
	{

		//load Configuration
		ConfigHandler::getInstance();
		//initialize NodeHandler
		if(ConfigHandler::loaded())
		{
			NodeHandler::getInstance();
			//initialize MapHandler
			MapHandler::getInstance();
			//initialize SpawnHandler
			SpawnHandler::getInstance();
			//initialize Traffic Lights
			TrafficLightHandler::getInstance();
			//after every numeric references to our streets have been solved, sort'em
			MapHandler::getInstance()->sortStreets();
			if(!MapHandler::loaded() || !NodeHandler::loaded() ||!SpawnHandler::loaded() || !TrafficLightHandler::loaded())
			{
				LOG->writeWorldLog("Unable to load world, one or more handlers couldn't be loaded. Please check the log for further information.");
				//don't close window immediatley
				std::string tmp;
				std::getline(std::cin, tmp);
			}
			else
			{
				_bLoaded = true;
			}
		}
	}

	std::string world::getInputFile(char fileType)
	{
		std::string prefix = "";
		switch(fileType)
		{
		case FILE_NODES:
			prefix = "nodes";
			break;
		case FILE_SPAWNS:
			prefix = "spawns";
			break;
		case FILE_INTERSECTIONS:
			prefix = "intersection";
			break;
		case FILE_CROSSROADS:
			prefix = "crossroads";
			break;
		case FILE_STREETS:
			prefix = "ways";
		}
		std::string rName = PATH_DATA;
#ifdef _WIN32
		rName = rName + INPUT_MAP + "\\" + prefix + "_" + INPUT_MAP + "_tmp.dat";
#else
		rName = rName + INPUT_MAP + "/" + prefix + "_" + INPUT_MAP + "_tmp.dat";
#endif
		return rName;
	}

	world::~world()
	{
		//free every assigned memory here
	}

	/// <summary>
	/// Returns the instance of our <see cref="NodeHandler" /> class to secure that only a single instance is created.
	/// </summary>
	/// <returns></returns>
	world* world::getInstance()
	{
		if(!_bInstance)
		{
			_instance = new world();
			_bInstance = true;
		}
		return _instance;
	}

	void world::update(boost::uint8_t step)
	{
		SpawnHandler::getInstance()->update();
		MSG->Synchronize();
		MapHandler::getInstance()->update();
		TrafficLightHandler::getInstance()->update();
		MSG->WaitMsgsSend();
		MSG->Synchronize();
		MapHandler::getInstance()->getCarHandler()->updateCarCount();
#ifdef TF_DEBUG_MAIN
		if(_lTimeStamp % 100 == 0 && world::getCommRank() == 0)
		{
			boost::format fmt = boost::format("%1% Cars on the map") % MapHandler::getInstance()->getCarHandler()->carCount();
			LOG->writeWorldLog(_FMT_TO_C_STR_(fmt));
		}
#endif
#ifdef CAR_DUMP
		if(_lTimeStamp % CAR_DUMP_STEP == 0)
			MAP->getCarHandler()->dumpCars();
#endif
		_lTimeStamp += step; //ncrease time stamp after reach loop

	}
}
